import numpy as np
import math
from scipy.spatial.distance import cdist


def eucl_dist(x, y):
    """
    Usage
    -----
    L2-norm between point x and y

    Parameters
    ----------
    param x : numpy_array
    param y : numpy_array

    Returns
    -------
    dist : float
           L2-norm between x and y
    """
    dist = np.linalg.norm(x - y)
    return dist


def eucl_dist_traj(t1, t2):
    """
    Usage
    -----
    Pairwise L2-norm between point of trajectories t1 and t2

    Parameters
    ----------
    param t1 : len(t1)x2 numpy_array
    param t2 : len(t1)x2 numpy_array

    Returns
    -------
    dist : float
           L2-norm between x and y
    """
    mdist = cdist(t1, t2, 'euclidean')
    return mdist


def point_to_seg(p, s1, s2, dps1, dps2, ds):
    """
    Usage
    -----
    Point to segment distance between point p and segment delimited by s1 and s2

    Parameters
    ----------
    param p : 1x2 numpy_array
    param s1 : 1x2 numpy_array
    param s2 : 1x2 numpy_array
    dps1 : euclidean distance between p and s1
    dps2 : euclidean distance between p and s2
    dps : euclidean distance between s1 and s2

    Returns
    -------
    dpl: float
         Point to segment distance between p and s
    """
    px = p[0]
    py = p[1]
    p1x = s1[0]
    p1y = s1[1]
    p2x = s2[0]
    p2y = s2[1]
    if p1x == p2x and p1y == p2y:
        dpl = dps1
    else:
        segl = ds
        x_diff = p2x - p1x
        y_diff = p2y - p1y
        u1 = (((px - p1x) * x_diff) + ((py - p1y) * y_diff))
        u = u1 / (segl * segl)

        if (u < 0.00001) or (u > 1):
            # closest point does not fall within the line segment, take the shorter distance to an endpoint
            dpl = min(dps1, dps2)
        else:
            # Intersecting point is on the line, use the formula
            ix = p1x + u * x_diff
            iy = p1y + u * y_diff
            dpl = eucl_dist(p, np.array([ix, iy]))

    return dpl


def point_to_trajectory(p, t, mdist_p, t_dist, l_t):
    """
    Usage
    -----
    Point-to-trajectory distance between point p and the trajectory t.
    The Point to trajectory distance is the minimum of point-to-segment distance between p and all segment s of t

    Parameters
    ----------
    param p: 1x2 numpy_array
    param t : l_tx2 numpy_array
    param mdist_p : l_t x 1 numpy array, distances from point p to points of trajectory t
    param t_dist : l_t x 1 numpy array, distances from consecutives points in trajectory t
    param l_t: int lenght of t

    Returns
    -------
    dpt : float,
          Point-to-trajectory distance between p and trajectory t
    """
    dpt = min(
        [point_to_seg(p, t[it], t[it + 1], mdist_p[it], mdist_p[it + 1], t_dist[it]) for it in range(l_t - 1)])
    return dpt


def circle_line_intersection(px, py, s1x, s1y, s2x, s2y, eps):
    """
    Usage
    -----
    Find the intersections between the circle of radius eps and center (px, py) and the line delimited by points
    (s1x, s1y) and (s2x, s2y).
    It is supposed here that the intersection between them exists. If no, raise error

    Parameters
    ----------
    param px : float, centre's abscissa of the circle
    param py : float, centre's ordinate of the circle
    param eps : float, radius of the circle
    param s1x : abscissa of the first point of the line
    param s1y : ordinate of the first point of the line
    param s2x : abscissa of the second point of the line
    param s2y : ordinate of the second point of the line

    Returns
    -------
    intersect : 2x2 numpy_array
                Coordinate of the two intersections.
                If the two intersections are the same, that's means that the line is a tangent of the circle.
    """
    if s2x == s1x:
        rac = math.sqrt((eps * eps) - ((s1x - px) * (s1x - px)))
        y1 = py + rac
        y2 = py - rac
        intersect = np.array([[s1x, y1], [s1x, y2]])
    else:
        m = (s2y - s1y) / (s2x - s1x)
        c = s2y - m * s2x
        A = m * m + 1
        B = 2 * (m * c - m * py - px)
        C = py * py - eps * eps + px * px - 2 * c * py + c * c
        delta = B * B - 4 * A * C
        if delta <= 0:
            x = -B / (2 * A)
            y = m * x + c
            intersect = np.array([[x, y], [x, y]])
        elif delta > 0:
            sdelta = math.sqrt(delta)
            x1 = (-B + sdelta) / (2 * A)
            y1 = m * x1 + c
            x2 = (-B - sdelta) / (2 * A)
            y2 = m * x2 + c
            intersect = np.array([[x1, y1], [x2, y2]])
        else:
            raise ValueError("The intersection between circle and line is supposed to exist")
    return intersect



#cython: boundscheck=False
#cython: wraparound=False

STUFF = "Hi"

from libc.math cimport sqrt
from libc.math cimport fmin
cimport numpy as np
import numpy as np
from cpython cimport bool



cdef double _eucl_dist(double x1,double y1,double x2,double y2):
    """
    Usage
    -----
    L2-norm between point (x1,y1) and (x2,y2)

    Parameters
    ----------
    param x1 : float
    param y1 : float
    param x2 : float
    param y2 : float

    Returns
    -------
    dist : float
           L2-norm between (x1,y1) and (x2,y2)
    """
    cdef double d,dx,dy
    dx=(x1-x2)
    dy=(y1-y2)
    dist=sqrt(dx*dx+dy*dy)
    return dist


def c_eucl_dist(double x1,double y1,double x2,double y2):
    """
    Usage
    -----
    L2-norm between point (x1,y1) and (x2,y2)

    Parameters
    ----------
    param x1 : float
    param y1 : float
    param x2 : float
    param y2 : float

    Returns
    -------
    dist : float
           L2-norm between (x1,y1) and (x2,y2)
    """
    cdef double d,dx,dy
    dx=(x1-x2)
    dy=(y1-y2)
    d=sqrt(dx*dx+dy*dy)
    return d



cdef double _point_to_seg(double px, double py, double p1x, double p1y, double p2x, double p2y ):
    """
    Usage
    -----
    Point to segment distance between point (px, py) and segment delimited by (p1x, p1y) and (p2x, p2y)

    Parameters
    ----------
    param px : float, abscissa of the point
    param py : float, ordinate of the point
    param p1x : float, abscissa of the first end point of the segment
    param p1y : float, ordinate of the first end point of the segment
    param p2x : float, abscissa of the second end point of the segment
    param p2y : float, ordinate of the second end point of the segment


    Returns
    -------
    dpl: float
         Point to segment distance between p and s
    """
    cdef double segl,u,ix,iy,dpl, dpx,dpy
    if p1x==p2x and p1y==p2y:
        dpl=_eucl_dist(px,py,p1x,p1y)
    else:
        segl= _eucl_dist(p1x,p1y,p2x,p2y)
        dpx = p2x - p1x
        dpy = p2y - p1y
        u = (((px - p1x) * dpx) + ((py - p1y) * dpy)) / (segl * segl)

        if (u < 0.00001) or (u > 1):
            #// closest point does not fall within the line segment, take the shorter distance
            #// to an endpoint
            ix = _eucl_dist(px,py,p1x,p1y)
            iy = _eucl_dist(px,py,p2x,p2y)
            dpl = fmin(ix,iy)
        else:
            # Intersecting point is on the line, use the formula
            ix = p1x + u * dpx
            iy = p1y + u * dpy
            dpl = _eucl_dist(px,py,ix,iy)

    return dpl

def c_point_to_seg(double px, double py, double p1x, double p1y, double p2x, double p2y ):
    """
    Usage
    -----
    Point to segment distance between point (px, py) and segment delimited by (p1x, p1y) and (p2x, p2y)

    Parameters
    ----------
    param px : float, abscissa of the point
    param py : float, ordinate of the point
    param p1x : float, abscissa of the first end point of the segment
    param p1y : float, ordinate of the first end point of the segment
    param p2x : float, abscissa of the second end point of the segment
    param p2y : float, ordinate of the second end point of the segment


    Returns
    -------
    dpl: float
         Point to segment distance between p and s
    """
    cdef double segl,u,ix,iy,dpl, dpx,dpy
    if p1x==p2x and p1y==p2y:
        dpl=_eucl_dist(px,py,p1x,p1y)
    else:
        segl= _eucl_dist(p1x,p1y,p2x,p2y)
        dpx = p2x - p1x
        dpy = p2y - p1y
        u = (((px - p1x) * dpx) + ((py - p1y) * dpy)) / (segl * segl)

        if (u < 0.00001) or (u > 1):
            #// closest point does not fall within the line segment, take the shorter distance
            #// to an endpoint
            ix = _eucl_dist(px,py,p1x,p1y)
            iy = _eucl_dist(px,py,p2x,p2y)
            dpl = fmin(ix,iy)
        else:
            # Intersecting point is on the line, use the formula
            ix = p1x + u * dpx
            iy = p1y + u * dpy
            dpl = _eucl_dist(px,py,ix,iy)

    return dpl

cdef double _point_to_trajectory (double px,double py, np.ndarray[np.float64_t,ndim=2] t):
    """
    Usage
    -----
    Point-to-trajectory distance between point (px, py) and the trajectory t.
    The Point to trajectory distance is the minimum of point-to-segment distance between p and all segment s of t

    Parameters
    ----------
    param px : float, abscissa of the point
    param py : float, ordinate of the point
    param t : len(t)x2 numpy_array

    Returns
    -------
    dpt : float,
          Point-to-trajectory distance between p and trajectory t
    """
    cdef double dpt,dist_j0
    cdef int nt,i
    cdef np.ndarray[np.float64_t,ndim=1] xt,yt

    nt=len(t)
    xt=t[:,0]
    yt=t[:,1]
    dist_j0=9e100
    for i from 0 <= i < (nt-1):
        dist_j0=fmin(dist_j0,_point_to_seg(px,py,xt[i],yt[i],xt[i+1],yt[i+1]))

    return dist_j0

def c_point_to_trajectory (double px,double py, np.ndarray[np.float64_t,ndim=2] t):
    """
    Usage
    -----
    Point-to-trajectory distance between point (px, py) and the trajectory t.
    The Point to trajectory distance is the minimum of point-to-segment distance between p and all segment s of t

    Parameters
    ----------
    param px : float, abscissa of the point
    param py : float, ordinate of the point
    param t : len(t)x2 numpy_array

    Returns
    -------
    dpt : float,
          Point-to-trajectory distance between p and trajectory t
    """
    cdef double dpt,dist_j0
    cdef int nt,i
    cdef np.ndarray[np.float64_t,ndim=1] xt,yt

    nt=len(t)
    xt=t[:,0]
    yt=t[:,1]
    dist_j0=9e100
    for i from 0 <= i < (nt-1):
        dist_j0=fmin(dist_j0,_point_to_seg(px,py,xt[i],yt[i],xt[i+1],yt[i+1]))

    return dist_j0

cdef np.ndarray[np.float64_t,ndim=2] _circle_line_intersection(double px, double py, double s1x, double s1y,
double
s2x, double s2y, double eps):
    """
    Usage
    -----
    Find the intersections between the circle of radius eps and center (px, py) and the line delimited by points
    (s1x, s1y) and (s2x, s2y).
    It is supposed here that the intersection between them exists. If no, raise error

    Parameters
    ----------
    param px : float, centre's abscissa of the circle
    param py : float, centre's ordinate of the circle
    eps : float, radius of the circle
    s1x : abscissa of the first point of the line
    s1y : ordinate of the first point of the line
    s2x : abscissa of the second point of the line
    s2y : ordinate of the second point of the line

    Returns
    -------
    intersect : 2x2 numpy_array
                Coordinate of the two intersections.
                If the two intersections are the same, that's means that the line is a tangent of the circle.
    """
    cdef double rac,y1,y2,m,c,A,B,C,delta,x,y,sdelta,x1,x2
    cdef np.ndarray[np.float64_t,ndim=2] intersect
    cdef bool delta_strict_inf_0,delta_inf_0,delta_strict_sup_0

    if s2x==s1x :
        rac=sqrt((eps*eps) - ((s1x-px)*(s1x-px)))
        y1 = py+rac
        y2 = py-rac
        intersect = np.array([[s1x,y1],[s1x,y2]])
    else:
        m= (s2y-s1y)/(s2x-s1x)
        c= s2y-m*s2x
        A=m*m+1
        B=2*(m*c-m*py-px)
        C=py*py-eps*eps+px*px-2*c*py+c*c
        delta=B*B-4*A*C
        delta_inf_0 = delta <= 0
        delta_strict_sup_0 = delta > 0
        if delta_inf_0 :
            delta_strict_inf_0 = delta <0
            x = -B/(2*A)
            y = m*x+c
            intersect = np.array([[x,y],[x,y]])
        elif delta_strict_sup_0 > 0 :
            sdelta = sqrt(delta)
            x1= (-B+sdelta)/(2*A)
            y1=m*x1+c
            x2= (-B-sdelta)/(2*A)
            y2=m*x2+c
            intersect = np.array([[x1,y1],[x2,y2]])
        else :
            raise ValueError("The intersection between circle and line is supposed to exist")
    return intersect

def c_circle_line_intersection(double px, double py, double s1x, double s1y, double
s2x, double s2y, double eps):
    """
    Usage
    -----
    Find the intersections between the circle of radius eps and center (px, py) and the line delimited by points
    (s1x, s1y) and (s2x, s2y).
    It is supposed here that the intersection between them exists. If no, raise error

    Parameters
    ----------
    param px : float, centre's abscissa of the circle
    param py : float, centre's ordinate of the circle
    param eps : float, radius of the circle
    param s1x : abscissa of the first point of the line
    param s1y : ordinate of the first point of the line
    param s2x : abscissa of the second point of the line
    param s2y : ordinate of the second point of the line

    Returns
    -------
    intersect : 2x2 numpy_array
                Coordinate of the two intersections.
                If the two intersections are the same, that's means that the line is a tangent of the circle.
    """
    cdef double rac,y1,y2,m,c,A,B,C,delta,x,y,sdelta,x1,x2
    cdef np.ndarray[np.float64_t,ndim=2] intersect
    cdef bool delta_strict_inf_0,delta_inf_0,delta_strict_sup_0

    if s2x==s1x :
        rac=sqrt((eps*eps) - ((s1x-px)*(s1x-px)))
        y1 = py+rac
        y2 = py-rac
        intersect = np.array([[s1x,y1],[s1x,y2]])
    else:
        m= (s2y-s1y)/(s2x-s1x)
        c= s2y-m*s2x
        A=m*m+1
        B=2*(m*c-m*py-px)
        C=py*py-eps*eps+px*px-2*c*py+c*c
        delta=B*B-4*A*C
        delta_inf_0 = delta <= 0
        delta_strict_sup_0 = delta > 0
        if delta_inf_0 :
            delta_strict_inf_0 = delta <0
            x = -B/(2*A)
            y = m*x+c
            intersect = np.array([[x,y],[x,y]])
        elif delta_strict_sup_0 > 0 :
            sdelta = sqrt(delta)
            x1= (-B+sdelta)/(2*A)
            y1=m*x1+c
            x2= (-B-sdelta)/(2*A)
            y2=m*x2+c
            intersect = np.array([[x1,y1],[x2,y2]])
        else :
            raise ValueError("The intersection between circle and line is supposed to exist")
    return intersect


STUFF = "Hi"

from libc.math cimport acos
from libc.math cimport sin
from libc.math cimport fmin
from libc.math cimport fmax
from libc.math cimport sqrt

from basic_euclidean import c_eucl_dist
cimport numpy as np
import numpy as np

cdef float PI = 3.14159265
cdef float HPI = 3.14159265/2

cdef double _ordered_mixed_distance(double six,double siy,double eix,double eiy,double sjx,double sjy,double ejx,
                                    double ejy,double sieix,double sieiy,double sjejx,double sjejy,
                                    double siei_norm_2,double sjej_norm_2):

    cdef double siei_norm,sjej_norm,sisjx,sisjy,siejx,siejy,u1,u2,psx,psy,pex,pey,cos_theta

    siei_norm=sqrt(siei_norm_2)
    sjej_norm=sqrt(sjej_norm_2)
    sisjx=sjx-six
    sisjy=sjy-siy
    siejx=ejx-six
    siejy=ejy-siy


    u1=(sisjx*sieix+sisjy*sieiy)/siei_norm_2
    u2=(siejx*sieix+siejy*sieiy)/siei_norm_2

    psx=six+u1*sieix
    psy=siy+u1*sieiy
    pex=six+u2*sieix
    pey=siy+u2*sieiy

    cos_theta = fmax(-1,fmin(1,(sjejx*sieix+sjejy*sieiy)/(siei_norm*sjej_norm)))
    theta = acos(cos_theta)

    #perpendicular distance
    lpe1=c_eucl_dist(sjx,sjy,psx,psy)
    lpe2=c_eucl_dist(ejx,ejy,pex,pey)
    if lpe1==0 and lpe2==0:
        dped= 0
    else:
        dped = (lpe1*lpe1+lpe2*lpe2)/(lpe1+lpe2)


    #parallel_distance
    lpa1=fmin(c_eucl_dist(six,siy,psx,psy),c_eucl_dist(eix,eiy,psx,psy))
    lpa2=fmin(c_eucl_dist(six,siy,pex,pey),c_eucl_dist(eix,eiy,pex,pey))
    dpad=fmin(lpa1,lpa2)

    #angle_distance
    if 0 <= theta <HPI :
        dad=sjej_norm * sin(theta)
    elif HPI <= theta <= PI:
        dad=sjej_norm
    else:
        raise ValueError("WRONG THETA")
    fdist = (dped+dpad+dad)/3

    return fdist


cdef double _mixed_distance(double six,double siy,double eix,double eiy,double sjx,double sjy,double ejx,double ejy):

    cdef double sieix,sieiy,sjejx,sjejy,siei_norm_2,sjej_norm_2,md

    sieix=eix-six
    sieiy=eiy-siy
    sjejx=ejx-sjx
    sjejy=ejy-sjy


    siei_norm_2=(sieix*sieix)+(sieiy*sieiy)
    sjej_norm_2=(sjejx*sjejx)+(sjejy*sjejy)

    if sjej_norm_2 > siei_norm_2 :
        md=_ordered_mixed_distance(sjx,sjy,ejx,ejy,six,siy,eix,eiy,sjejx,sjejy,sieix,sieiy,sjej_norm_2,siei_norm_2)
    else :
        md=_ordered_mixed_distance(six,siy,eix,eiy,sjx,sjy,ejx,ejy,sieix,sieiy,sjejx,sjejy,siei_norm_2,sjej_norm_2)

    return md

def c_segments_distance(np.ndarray[np.float64_t,ndim=2] traj_0,np.ndarray[np.float64_t,ndim=2] traj_1):

    cdef np.ndarray[np.float64_t,ndim=2] M
    cdef int n0,n1,i,j


    n0=len(traj_0)
    n1=len(traj_1)
    M=np.zeros((n0-1,n1-1))
    for i in range(n0-1):
        for j in range(n1-1):
            M[i,j]=_mixed_distance(traj_0[i,0],traj_0[i,1],traj_0[i+1,0],traj_0[i+1,1],
                                   traj_1[j,0],traj_1[j,1],traj_1[j+1,0],traj_1[j+1,1])
    return M








from .pydist.linecell import trajectory_set_grid

from .cydist.sspd import c_e_sspd, c_g_sspd
from .cydist.dtw import c_e_dtw, c_g_dtw
from .cydist.erp import c_e_erp, c_g_erp
from .cydist.edr import c_e_edr, c_g_edr
from .cydist.lcss import c_e_lcss, c_g_lcss
from .cydist.hausdorff import c_e_hausdorff, c_g_hausdorff
from .cydist.discret_frechet import c_discret_frechet
from .cydist.frechet import c_frechet
from .cydist.sowd import c_sowd_grid

import numpy as np

import warnings

__all__ = ["pdist", "cdist", "sspd", "sowd_grid", "frechet", "discret_frechet", "hausdorff", "dtw", "lcss", "edr",
           "erp"]

METRIC_DIC = {"spherical": {"sspd": c_g_sspd,
                            "dtw": c_g_dtw,
                            "lcss": c_g_lcss,
                            "hausdorff": c_g_hausdorff,
                            "sowd_grid": c_sowd_grid,
                            "erp": c_g_erp,
                            "edr": c_g_edr},
              "euclidean": {"sspd": c_e_sspd,
                            "dtw": c_e_dtw,
                            "lcss": c_e_lcss,
                            "hausdorff": c_e_hausdorff,
                            "discret_frechet": c_discret_frechet,
                            "frechet": c_frechet,
                            "sowd_grid": c_sowd_grid,
                            "erp": c_e_erp,
                            "edr": c_e_edr}}


# #################
# Simple Distance #
# #################

def sspd(traj_1, traj_2, type_d="euclidean"):
    """
    Usage
    -----
    Compute the Symmetrized Segment-Path Distance between trajectory traj_1 and traj_2

    type_d parameter options are "euclidean" or "spherical".
    - "euclidean" can handle every types of coordinate.
    - "spherical" implies a 2-dimensional trajectory where first dimensions is longitudes coordinate and second
    dimension is latitude coordinate.

    Parameters
    ----------
    param traj_1         : a 2-D numpy array
    param traj_2         : a 2-D numpy array
    param type_d         : string, distance type_d used (spherical or euclidean)

    Returns
    -------
    dist : a float. The sspd distance between traj_1 and traj_2
    """
    dim_1 = traj_1.shape[1]
    dim_2 = traj_2.shape[1]

    if dim_1 != 2 or dim_2 != 2:
        raise ValueError("Trajectories should be in 2D. t1 is %dD and t2 is %d given" % (dim_1, dim_2))

    if not (type_d in ["spherical", "euclidean"]):
        raise ValueError("The type_d argument should be 'euclidean' or 'spherical'\ntype_d given is : " + type_d)

    dist_func = METRIC_DIC[type_d]["sspd"]
    dist = dist_func(traj_1, traj_2)
    return dist


def sowd_grid(traj_1, traj_2, type_d="euclidean", converted=None, precision=None):
    """
    Usage
    -----
    Compute the Symmetrized One_Way Distance between trajectory traj_1 and traj_2

    type_d's parameters available are "euclidean" or "spherical".
    - "euclidean" can handle every types of coordinate.
    - "spherical" implies a 2-dimensional trajectory where first dimensions is longitudes coordinate and second
    dimension is latitute coordinate.

    'sowd_grid' computes distance between trajectory in grid representation. If the coordinate
    are spherical, this conversion can be made according to the geohash encoding. If so, the geohash 'precision'
    is needed. Otherwise trajectory are considered encoded in cells representation


    Parameters
    ----------

    param traj_1:        : a 2-D numpry array
    param traj_2:        : a 2-D numpry array
    param type_d :       : string, distance type_d used (spherical or euclidean)
    param converted      : Boolean, if True trajectories are considered in cells representation
    param precision      : Geohash precisions for conversion

    Returns
    -------

    dist : a float. The sowd_grid distance between traj_1 and traj_2
    """
    dim_1 = traj_1.shape[1]
    dim_2 = traj_2.shape[1]

    if dim_1 != 2 or dim_2 != 2:
        raise ValueError("Trajectories should be in 2D. t1 is %dD and t2 is %d given" % (dim_1, dim_2))

    if not (type_d in ["spherical", "euclidean"]):
        raise ValueError("The type_d argument should be 'euclidean' or 'spherical'\ntype_d given is : " + type_d)

    if type_d == "euclidean" and not converted:
        raise Warning("Euclidean implementation for distance sowd_grid is not "
                      "disponible if your data is not already converted in cell format")

    if converted is None:
        warnings.warn("converted parameter should be specified for metric sowd_grid. Default "
                      "is True")
        converted = True
    if converted:
        cells_list = [traj_1, traj_2]
    else:
        if precision is None:
            warnings.warn("precision parameter should be specified for metric sowd_grid if converted "
                          "is False. Default is 7")
            precision = 7
        cells_list_, _, _, _, _ = trajectory_set_grid([traj_1, traj_2], precision)
        cells_list = [np.array(cells_list_[0])[:, :2], np.array(cells_list_[1])[:, :2]]

    dist_func = METRIC_DIC[type_d]["sowd_grid"]
    dist = dist_func(cells_list[0], cells_list[1])
    return dist


def frechet(traj_1, traj_2, type_d="euclidean"):
    """
    Usage
    -----
    Compute the Frechet Distance between trajectory traj_1 and traj_2

    type_d's parameter available is "euclidean".


    Parameters
    ----------

    param traj_1         : a 2-D numpry array
    param traj_2         : a 2-D numpry array
    param type_d         : string, distance type_d used (spherical or euclidean)
    

    Returns
    -------

    dist : a float. The frechet distance between traj_1 and traj_2
    """
    dim_1 = traj_1.shape[1]
    dim_2 = traj_2.shape[1]

    if dim_1 != 2 or dim_2 != 2:
        raise ValueError("Trajectories should be in 2D. t1 is %dD and t2 is %d given" % (dim_1, dim_2))

    if type_d != "euclidean":
        raise ValueError("The type_d argument should be 'euclidean'\ntype_d given is : " + type_d)

    dist_func = METRIC_DIC[type_d]["frechet"]
    dist = dist_func(traj_1, traj_2)
    return dist


def discret_frechet(traj_1, traj_2, type_d="euclidean"):
    """
    Usage
    -----
    Compute the Discrete Frechet Distance between trajectory traj_1 and traj_2

    type_d's parameter available is "euclidean".

    Parameters
    ----------

    param traj_1         : a 2-D numpry array
    param traj_2         : a 2-D numpry array
    param type_d         : string, distance type_d used (spherical or euclidean)
    

    Returns
    -------

    dist : a float. The discret_frechet distance between traj_1 and traj_2
    """
    dim_1 = traj_1.shape[1]
    dim_2 = traj_2.shape[1]

    if dim_1 != 2 or dim_2 != 2:
        raise ValueError("Trajectories should be in 2D. t1 is %dD and t2 is %d given" % (dim_1, dim_2))

    if type_d != "euclidean":
        raise ValueError("The type_d argument should be 'euclidean'\ntype_d given is : " + type_d)

    dist_func = METRIC_DIC[type_d]["discret_frechet"]
    dist = dist_func(traj_1, traj_2)
    return dist


def hausdorff(traj_1, traj_2, type_d="euclidean"):
    """
    Usage
    -----
    Compute the Hausdorff Distance between trajectory traj_1 and traj_2

    type_d's parameters available are "euclidean" or "spherical".
    - "euclidean" can handle every types of coordinate.
    - "spherical" implies a 2-dimensional trajectory where first dimensions is longitudes coordinate and second
    dimension is latitute coordinate.

    If the distance traj_0 and traj_1 are 2-dimensional, the cython implementation is used, otherwised the python one
    is used, unless "python" implementation is specified


    Parameters
    ----------

    param traj_1         : a 2-D numpry array
    param traj_2         : a 2-D numpry array
    param type_d         : string, distance type_d used (spherical or euclidean)
    

    Returns
    -------

    dist : a float. The hausdorff distance between traj_1 and traj_2
    """
    dim_1 = traj_1.shape[1]
    dim_2 = traj_2.shape[1]

    if dim_1 != 2 or dim_2 != 2:
        raise ValueError("Trajectories should be in 2D. t1 is %dD and t2 is %d given" % (dim_1, dim_2))

    if not (type_d in ["spherical", "euclidean"]):
        raise ValueError("The type_d argument should be 'euclidean' or 'spherical'\ntype_d given is : " + type_d)

    dist_func = METRIC_DIC[type_d]["hausdorff"]
    dist = dist_func(traj_1, traj_2)
    return dist


def dtw(traj_1, traj_2, type_d="euclidean"):
    """
    Usage
    -----
    Compute the Dynamic Time Warping Distance between trajectory traj_1 and traj_2

    type_d's parameters available are "euclidean" or "spherical".
    - "euclidean" can handle every types of coordinate.
    - "spherical" implies a 2-dimensional trajectory where first dimensions is longitudes coordinate and second
    dimension is latitute coordinate.

    Parameters
    ----------

    param traj_1         : a 2-D numpry array
    param traj_2         : a 2-D numpry array
    param type_d         : string, distance type_d used (spherical or euclidean)
    

    Returns
    -------

    dist : a float. The dtw distance between traj_1 and traj_2
    """
    dim_1 = traj_1.shape[1]
    dim_2 = traj_2.shape[1]

    if dim_1 != 2 or dim_2 != 2:
        raise ValueError("Trajectories should be in 2D. t1 is %dD and t2 is %d given" % (dim_1, dim_2))

    if not (type_d in ["spherical", "euclidean"]):
        raise ValueError("The type_d argument should be 'euclidean' or 'spherical'\ntype_d given is : " + type_d)

    dist_func = METRIC_DIC[type_d]["dtw"]
    dist = dist_func(traj_1, traj_2)
    return dist


def lcss(traj_1, traj_2, type_d="euclidean", eps=200):
    """
    Usage
    -----
    Compute the Longuest Common Subsequence Distance between trajectory traj_1 and traj_2

    type_d's parameters available are "euclidean" or "spherical".
    - "euclidean" can handle every types of coordinate.
    - "spherical" implies a 2-dimensional trajectory where first dimensions is longitudes coordinate and second
    dimension is latitute coordinate.

    Parameters
    ----------

    param traj_1         : a 2-D numpry array
    param traj_2         : a 2-D numpry array
    param type_d         : string, distance type_d used (spherical or euclidean)
    

    Returns
    -------

    dist : a float. The lcss distance between traj_1 and traj_2
    """
    dim_1 = traj_1.shape[1]
    dim_2 = traj_2.shape[1]

    if dim_1 != 2 or dim_2 != 2:
        raise ValueError("Trajectories should be in 2D. t1 is %dD and t2 is %d given" % (dim_1, dim_2))

    if not (type_d in ["spherical", "euclidean"]):
        raise ValueError("The type_d argument should be 'euclidean' or 'spherical'\ntype_d given is : " + type_d)

    dist_func = METRIC_DIC[type_d]["lcss"]
    dist = dist_func(traj_1, traj_2, eps)
    return dist


def edr(traj_1, traj_2, type_d="euclidean", eps=200):
    """
    Usage
    -----
    Compute the Edit distance on Real Sequence between trajectory traj_1 and traj_2

    type_d's parameters available are "euclidean" or "spherical".
    - "euclidean" can handle every types of coordinate.
    - "spherical" implies a 2-dimensional trajectory where first dimensions is longitudes coordinate and second
    dimension is latitute coordinate.


    Parameters
    ----------

    param traj_1         : a 2-D numpry array
    param traj_2         : a 2-D numpry array
    param type_d         : string, distance type_d used (spherical or euclidean)
    

    Returns
    -------

    dist : a float. The edr distance between traj_1 and traj_2
    """
    dim_1 = traj_1.shape[1]
    dim_2 = traj_2.shape[1]

    if dim_1 != 2 or dim_2 != 2:
        raise ValueError("Trajectories should be in 2D. t1 is %dD and t2 is %d given" % (dim_1, dim_2))

    if not (type_d in ["spherical", "euclidean"]):
        raise ValueError("The type_d argument should be 'euclidean' or 'spherical'\ntype_d given is : " + type_d)

    dist_func = METRIC_DIC[type_d]["edr"]
    dist = dist_func(traj_1, traj_2, eps)
    return dist


def erp(traj_1, traj_2, type_d="euclidean", g=None):
    """
    Usage
    -----
    Compute theEdit Distance on Real Sequence between trajectory traj_1 and traj_2

    type_d's parameters available are "euclidean" or "spherical".
    - "euclidean" can handle every types of coordinate.
    - "spherical" implies a 2-dimensional trajectory where first dimensions is longitudes coordinate and second
    dimension is latitute coordinate.


    Parameters
    ----------

    param traj_1         : a 2-D numpry array
    param traj_2         : a 2-D numpry array
    param type_d         : string, distance type_d used (spherical or euclidean)
    

    Returns
    -------

    dist : a float. The erp distance between traj_1 and traj_2
    """
    dim_1 = traj_1.shape[1]
    dim_2 = traj_2.shape[1]

    if dim_1 != 2 or dim_2 != 2:
        raise ValueError("Trajectories should be in 2D. t1 is %dD and t2 is %d given" % (dim_1, dim_2))
    dim = dim_1

    if g is None:
        g = np.zeros(dim, dtype=float)
        warnings.warn("g parameter should be specified for metric erp. Default is ")
        print(g)
    else:
        if g.shape[0] != dim:
            raise ValueError("g and trajectories in list should have same dimension")

    if not (type_d in ["spherical", "euclidean"]):
        raise ValueError("The type_d argument should be 'euclidean' or 'spherical'\ntype_d given is : " + type_d)

    dist_func = METRIC_DIC[type_d]["erp"]
    dist = dist_func(traj_1, traj_2, g)
    return dist


# ####################
# Pairwise Distance #
# ####################

def pdist(traj_list, metric="sspd", type_d="euclidean", converted=None, precision=None,
          eps=None, g=None, verbose=False):
    """
    Usage
    -----
    Pairwise distances between trajectory in traj_list.

    metrics available are :

    1. 'sspd'

        Computes the distances using the Symmetrized Segment Path distance.

    2. 'dtw'

        Computes the distances using the Dynamic Path Warping distance.

    3. 'lcss'

        Computes the distances using the Longuest Common SubSequence distance

    4. 'hausdorf'

        Computes the distances using the Hausdorff distance.

    5. 'frechet'

        Computes the distances using the Frechet distance.

    6. 'discret_frechet'

        Computes the distances using the Discrete Frechet distance.

    7. 'sowd_grid'

        Computes the distances using the Symmetrized One Way Distance.

    8. 'erp'

        Computes the distances using the Edit Distance with real Penalty.

    9. 'edr'

        Computes the distances using the Edit Distance on Real sequence.

    type_d available are "euclidean" or "spherical". Some distance can be computing according to spherical space
    instead of euclidean. If so, traj_0 and traj_1 have to be 2-dimensional. First column is longitude, second one
    is latitude.

    'sowd_grid' computes distance between trajectory in grid representation. If the coordinate
    are spherical, this conversion can be made according to the geohash encoding. If so, the geohash 'precision'
    is needed.

    'edr' and 'lcss' require 'eps' parameter. These distance assume that two locations are similar, or not, according
    to a given threshold, eps.

    'erp' require g parameter. This distance require a gap parameter. Which must have same dimension that the
    trajectory.

    Parameters
    ----------

    param traj_list:       a list of nT numpy array trajectory
    param metric :         string, distance used
    param type_d :         string, distance type_d used (spherical or euclidean)
    
    param converted :      boolean, specified if the data are converted in cell format (sowd_grid
                           metric)
    param precision :      int, precision of geohash (sowd_grid )
    param eps :            float, threshold distance (edr and lcss)
    param g :              numpy arrays, gaps (erp distance)


    Returns
    -------

    M : a nT x nT numpy array. Where the i,j entry is the distance between traj_list[i] and traj_list[j]
    """

    list_dim = [x.shape[1] if len(x.shape) > 1 else 1 for x in traj_list]
    nb_traj = len(traj_list)
    if not (len(set(list_dim)) == 1):
        raise ValueError("All trajectories must have same dimesion !")
    dim = list_dim[0]

    if not (metric in ["sspd", "dtw", "lcss", "hausdorff", "frechet", "discret_frechet", "sowd_grid",
                       "erp", "edr"]):
        raise ValueError("The metric argument should be 'sspd', 'dtw', 'lcss','erp','edr' 'hausdorff', 'frechet',"
                         "'discret_frechet' or 'sowd_grid' \nmetric given is : " + metric)

    if not (type_d in ["spherical", "euclidean"]):
        raise ValueError("The type_d argument should be 'euclidean' or 'spherical'\ntype_d given is : " + type_d)

    if type_d == "spherical" and (metric in ["frechet", "discret_frechet"]):
        raise ValueError("spherical implementation for distance " + metric + " is not "
                                                                             "disponible")
    if type_d == "euclidean" and (metric in ["sowd", "sowd_grid"]):
        if not (converted):
            raise ValueError("Euclidean implementation for distance " +
                             metric + " is not disponible if your data is not already converted in cell format")

    if verbose:
       print(("Computing " + type_d + " distance " + metric + " for %d trajectories" % nb_traj))
    M = np.zeros(sum(range(nb_traj)))
    dist = METRIC_DIC[type_d][metric]
    if metric.startswith("sowd_grid"):
        if converted is None:
            warnings.warn("converted parameter should be specified for metric sowd_grid. Default "
                          "is False")
            converted = False
        if converted:
            cells_list = traj_list
        else:
            if precision is None:
                warnings.warn("precision parameter should be specified for metric sowd_grid if converted "
                              "is False. Default is 7")
                precision = 7
            if verbose:
                print("Cells conversion start")
            cells_list_, _, _, _, _ = trajectory_set_grid(traj_list, precision)
            cells_list = [np.array(x)[:, :2] for x in cells_list_]
            if verbose:
                print("Cells conversion ok")
        im = 0
        for i in range(nb_traj):
            cells_list_i = cells_list[i]
            for j in range(i + 1, nb_traj):
                cells_list_j = cells_list[j]
                M[im] = dist(cells_list_i, cells_list_j)
                im += 1
    elif metric == "erp":
        if g is None:
            g = np.zeros(dim, dtype=float)
            warnings.warn("g parameter should be specified for metric erp. Default is ")
            if verbose:
                print(g)
        else:
            if g.shape[0] != dim:
                raise ValueError("g and trajectories in list should have same dimension")
        im = 0
        for i in range(nb_traj):
            traj_list_i = traj_list[i]
            for j in range(i + 1, nb_traj):
                traj_list_j = traj_list[j]
                M[im] = dist(traj_list_i, traj_list_j, g)
                im += 1
    elif metric == "lcss" or metric == "edr":
        if eps is None:
            warnings.warn("eps parameter should be specified for metric 'lcss' and 'edr', default is 100 ")
            eps = 100
        im = 0
        for i in range(nb_traj):
            traj_list_i = traj_list[i]
            for j in range(i + 1, nb_traj):
                traj_list_j = traj_list[j]
                M[im] = dist(traj_list_i, traj_list_j, eps)
                im += 1
    else:
        im = 0
        for i in range(nb_traj):
            traj_list_i = traj_list[i]
            for j in range(i + 1, nb_traj):
                traj_list_j = traj_list[j]
                M[im] = dist(traj_list_i, traj_list_j)
                im += 1
    return M


# ########################
#  Distance between list #
# ########################

def cdist(traj_list_1, traj_list_2, metric="sspd", type_d="euclidean", converted=None,
          precision=None, eps=None, g=None, verbose=False):
    """
    Usage
    -----
    Computes distance between each pair of the two list of trajectories

    metrics available are :

    1. 'sspd'

        Computes the distances using the Symmetrized Segment Path distance.

    2. 'dtw'

        Computes the distances using the Dynamic Path Warping distance.

    3. 'lcss'

        Computes the distances using the Longuest Common SubSequence distance

    4. 'hausdorf'

        Computes the distances using the Hausdorff distance.

    5. 'frechet'

        Computes the distances using the Frechet distance.

    6. 'discret_frechet'

        Computes the distances using the Discrete Frechet distance.

    7. 'sowd_grid'

        Computes the distances using the Symmetrized One Way Distance.

    8. 'erp'

        Computes the distances using the Edit Distance with real Penalty.

    9. 'edr'

        Computes the distances using the Edit Distance on Real sequence.

    type_d available are "euclidean" or "spherical". Some distance can be computing according to spherical space
    instead of euclidean. If so, traj_0 and traj_1 have to be 2-dimensional. First column is longitude, second one
    is latitude.


    'sowd_grid', compute distance between trajectory in grid representation. If the coordinate
    are spherical, this conversion can be made according to the geohash encoding. If so, the geohash 'precision'
    is needed.

    'edr' and 'lcss' require 'eps' parameter. These distance assume that two locations are similar, or not, according
    to a given threshold, eps.

    'erp' require g parameter. This distance require a gap parameter. Which must have same dimension that the
    trajectory.

    Parameters
    ----------

    param traj_list:       a list of nT numpy array trajectory
    param metric :         string, distance used
    param type_d :         string, distance type_d used (spherical or euclidean)
    
    param converted :      boolean, specified if the data are converted in cell format (sowd_grid)
                           metric)
    param precision :      int, precision of geohash (sowd_grid)
    param eps :            float, threshold distance (edr and lcss)
    param g :              numpy arrays, gaps (erp distance)


    Returns
    -------

    M : a nT1 x nT2 numpy array. Where the i,j entry is the distance between traj_list_1[i] and traj_list_2[j]

    """

    list_dim_1 = [x.shape[1] for x in traj_list_1]
    nb_traj_1 = len(traj_list_1)
    list_dim_2 = [x.shape[1] for x in traj_list_2]
    nb_traj_2 = len(traj_list_2)
    if not (len(set(list_dim_1 + list_dim_2)) == 1):
        raise ValueError("All trajectories must have same dimesion !")
    dim = list_dim_1[0]

    if not (metric in ["sspd", "dtw", "lcss", "hausdorff", "frechet", "discret_frechet", "sowd_grid",
                       "erp", "edr"]):
        raise ValueError("The metric argument should be 'sspd', 'dtw', 'lcss','erp','edr' 'hausdorff', 'frechet',"
                         "'discret_frechet' or 'sowd_grid'\nmetric given is : " + metric)

    if not (type_d in ["spherical", "euclidean"]):
        raise ValueError("The type_d argument should be 'euclidean' or 'spherical'\ntype_d given is : " + type_d)

    if verbose:
        print(("Computing " + type_d + " distance " + metric + " for %d and %d trajectories" % (nb_traj_1, nb_traj_2)))
    M = np.zeros((nb_traj_1, nb_traj_2))
    dist = METRIC_DIC[type_d][metric]
    if metric.startswith("sowd_grid"):
        if converted is None:
            warnings.warn("converted parameter should be specified for metric sowd_grid. Default "
                          "is True")
            converted = True
        if converted:
            cells_list_1 = traj_list_1
            cells_list_2 = traj_list_2
        else:
            if precision is None:
                warnings.warn("precision parameter should be specified for metric sowd_grid  if converted "
                              "is False. Default is 7")
                precision = 7
            cells_list, _, _, _, _ = trajectory_set_grid(traj_list_1 + traj_list_2, precision)
            cells_list_1 = [np.array(x)[:, :2] for x in cells_list[:nb_traj_1]]
            cells_list_2 = [np.array(x)[:, :2] for x in cells_list[nb_traj_1:]]
        for i in range(nb_traj_1):
            cells_list_1_i = cells_list_1[i]
            for j in range(nb_traj_2):
                cells_list_2_j = cells_list_2[j]
                M[i, j] = dist(cells_list_1_i, cells_list_2_j)
    elif metric == "erp":
        if g is None:
            g = np.zeros(dim, dtype=float)
            warnings.warn("g parameter should be specified for metric erp. Default is ")
            if verbose:
                print(g)
        else:
            if g.shape[0] != dim:
                raise ValueError("g and trajectories in list should have same dimension")
        for i in range(nb_traj_1):
            traj_list_1_i = traj_list_1[i]
            for j in range(nb_traj_2):
                traj_list_2_j = traj_list_2[j]
                M[i, j] = dist(traj_list_1_i, traj_list_2_j, g)
    elif metric == "lcss" or metric == "edr":
        if eps is None:
            warnings.warn("eps parameter should be specified for metric 'lcss' and 'edr', default is 100 ")
            eps = 100
        for i in range(nb_traj_1):
            traj_list_1_i = traj_list_1[i]
            for j in range(nb_traj_2):
                traj_list_2_j = traj_list_2[j]
                M[i, j] = dist(traj_list_1_i, traj_list_2_j, eps)
    else:
        for i in range(nb_traj_1):
            traj_list_1_i = traj_list_1[i]
            for j in range(nb_traj_2):
                traj_list_2_j = traj_list_2[j]
                M[i, j] = dist(traj_list_1_i, traj_list_2_j)
    return M



STUFF = "Hi"

from libc.math cimport fmax
from libc.math cimport fmin


cimport numpy as np
import numpy as np


from basic_euclidean import c_eucl_dist
from basic_geographical import c_great_circle_distance

def c_e_lcss(np.ndarray[np.float64_t,ndim=2] t0, np.ndarray[np.float64_t,ndim=2] t1, eps):
    """
    Usage
    -----
    The Longuest-Common-Subsequence distance between trajectory t0 and t1.

    Parameters
    ----------
    param t0 : len(t0)x2 numpy_array
    param t1 : len(t1)x2 numpy_array
    eps : float

    Returns
    -------
    lcss : float
           The Longuest-Common-Subsequence distance between trajectory t0 and t1
    """
    cdef int n0,n1,i,j
    cdef np.ndarray[np.float64_t,ndim=2] C
    cdef double x0,y0,x1,y1,lcss

    n0 = len(t0)+1
    n1 = len(t1)+1

    # An (m+1) times (n+1) matrix
    C=np.zeros((n0,n1))

    for i from 1 <= i < n0:
        for j from 1 <= j < n1:
            x0=t0[i-1,0]
            y0=t0[i-1,1]
            x1=t1[j-1,0]
            y1=t1[j-1,1]
            if c_eucl_dist(x0,y0,x1,y1)<eps:
                C[i,j] = C[i-1,j-1] + 1
            else:
                C[i,j] = fmax(C[i,j-1], C[i-1,j])
    lcss = 1-float(C[n0-1,n1-1])/fmin(n0-1,n1-1)
    return lcss



def c_g_lcss(np.ndarray[np.float64_t,ndim=2] t0, np.ndarray[np.float64_t,ndim=2] t1, eps):
    """
    Usage
    -----
    The Longuest-Common-Subsequence distance between trajectory t0 and t1.

    Parameters
    ----------
    param t0 : len(t0)x2 numpy_array
    param t1 : len(t1)x2 numpy_array
    eps : float

    Returns
    -------
    lcss : float
           The Longuest-Common-Subsequence distance between trajectory t0 and t1
    """
    cdef int n0,n1,i,j
    cdef np.ndarray[np.float64_t,ndim=2] C
    cdef double x0,y0,x1,y1,lcss

    n0 = len(t0)+1
    n1 = len(t1)+1

    # An (m+1) times (n+1) matrix
    C=np.zeros((n0,n1))

    for i from 1 <= i < n0:
        for j from 1 <= j < n1:
            x0=t0[i-1,0]
            y0=t0[i-1,1]
            x1=t1[j-1,0]
            y1=t1[j-1,1]
            if c_great_circle_distance(x0,y0,x1,y1)<eps:
                C[i,j] = C[i-1,j-1] + 1
            else:
                C[i,j] = fmax(C[i,j-1], C[i-1,j])
    lcss = 1-float(C[n0-1,n1-1])/fmin(n0-1,n1-1)
    return lcss



import numpy as np
import traj_dist.distance as tdist
import pickle

traj_list = pickle.load(open("/Users/bguillouet/These/trajectory_distance/data/benchmark_trajectories.pkl", "rb"))[:10]
traj_A = traj_list[0]
traj_B = traj_list[1]



# Simple distance

dist = tdist.sspd(traj_A, traj_B)
print(dist)

# Pairwise distance

pdist = tdist.pdist(traj_list, metric="sspd")
print(pdist)

# Distance between two list of trajectories

cdist = tdist.cdist(traj_list, traj_list, metric="sspd")
print(cdist)



STUFF = "Hi"

from libc.math cimport sin
from libc.math cimport cos
from libc.math cimport acos
from libc.math cimport asin
from libc.math cimport atan2
from libc.math cimport sqrt
from libc.math cimport fmin
from libc.math cimport fabs

cimport numpy as np
import numpy as np

cdef float pi = 3.14159265
cdef float rad = pi/180.0
cdef int R = 6378137

cdef np.ndarray[np.float64_t,ndim=1] _spherical2Cart(double lon, double lat):

    cdef double clat,x,y,z

    clat=(90-lat)*rad
    lon=lon*rad
    x=cos(lon)*sin(clat)
    y=sin(lon)*sin(clat)
    z=cos(clat)

    return np.array([x,y,z])

def c_spherical2Cart(double lon, double lat):

    cdef double clat,x,y,z

    clat=(90-lat)*rad
    lon=lon*rad
    x=cos(lon)*sin(clat)
    y=sin(lon)*sin(clat)
    z=cos(clat)

    return np.array([x,y,z])

cdef np.ndarray[np.float64_t,ndim=1] _cart2Spherical(double x, double y, double  z):

    cdef r, clat, lat, lon

    r=sqrt(x**2+y**2+z**2)
    clat=acos(z/r)/pi*180
    lat=90.-clat
    lon=atan2(y,x)/pi*180
    lon=(lon+360)%360

    return np.array([lon,lat])

def c_cart2Spherical(double  x, double  y, double  z):

    cdef r, clat, lat, lon

    r=sqrt(x**2+y**2+z**2)
    clat=acos(z/r)/pi*180
    lat=90.-clat
    lon=atan2(y,x)/pi*180
    lon=(lon+360)%360

    return np.array([lon,lat])


cdef double _great_circle_distance(double lon1,double lat1,double lon2,double lat2):
    """
    Usage
    -----
    Compute the great circle distance, in meter, between (lon1,lat1) and (lon2,lat2)

    Parameters
    ----------
    param lat1: float, latitude of the first point
    param lon1: float, longitude of the first point
    param lat2: float, latitude of the second point
    param lon2: float, longitude of the second point

    Returns
    -------
    d: float
       Great circle distance between (lon1,lat1) and (lon2,lat2)
    """
    cdef double dLat,dLon,a,c,d

    dLat = (lat2 - lat1)*rad
    dLon = rad*(lon2 - lon1)
    a = (sin(dLat / 2) * sin(dLat / 2) +
    cos(rad*(lat1)) * cos(rad*(lat2)) *
    sin(dLon / 2) * sin(dLon / 2))
    c = 2 * atan2(sqrt(a), sqrt(1 - a))
    d = R * c
    return d


def c_great_circle_distance(double lon1,double lat1,double lon2,double lat2):
    """
    Usage
    -----
    Compute the great circle distance, in meter, between (lon1,lat1) and (lon2,lat2)

    Parameters
    ----------
    param lon1: float, longitude of the first point
    param lat1: float, latitude of the first point
    param lon2: float, longitude of the second point
    param lat2: float, latitude of the second point

    Returns
    -------
    d: float
       Great circle distance between (lon1,lat1) and (lon2,lat2)
    """
    cdef double dLat,dLon,a,c,d

    dLat = rad*(lat2 - lat1)
    dLon = rad*(lon2 - lon1)
    a = (sin(dLat / 2) * sin(dLat / 2) +
    cos(rad*(lat1)) * cos(rad*(lat2)) *
    sin(dLon / 2) * sin(dLon / 2))
    c = 2 * atan2(sqrt(a), sqrt(1 - a))
    d = R * c
    return d


cdef double _initial_bearing(double lon1,double lat1,double lon2,double lat2):
    """
    Usage
    -----
    Bearing between (lon1,lat1) and (lon2,lat2), in degree.

    Parameters
    ----------
    param lat1: float, latitude of the first point
    param lon1: float, longitude of the first point
    param lat2: float, latitude of the second point
    param lon2: float, longitude of the second point

    Returns
    -------
    brng: float
           Bearing between (lon1,lat1) and (lon2,lat2), in degree.

    """
    cdef double dLat,dLon,x,y,ibrng

    dLon = rad*(lon2 - lon1)
    y = sin(dLon) * cos(rad*(lat2))
    x = cos(rad*(lat1))*sin(rad*(lat2)) - sin(rad*(lat1))*cos(rad*(lat2))*cos(dLon)
    ibrng = atan2(y, x)

    return ibrng

def c_initial_bearing(double lon1,double lat1,double lon2,double lat2):
    """
    Usage
    -----
    Bearing between (lon1,lat1) and (lon2,lat2), in degree.

    Parameters
    ----------
    param lat1: float, latitude of the first point
    param lon1: float, longitude of the first point
    param lat2: float, latitude of the second point
    param lon2: float, longitude of the second point

    Returns
    -------
    brng: float
           Bearing between (lon1,lat1) and (lon2,lat2), in degree.

    """
    cdef double dLat,dLon,x,y,ibrng

    dLon = rad*(lon2 - lon1)
    y = sin(dLon) * cos(rad*(lat2))
    x = cos(rad*(lat1))*sin(rad*(lat2)) - sin(rad*(lat1))*cos(rad*(lat2))*cos(dLon)
    ibrng = atan2(y, x)

    return ibrng

cdef np.ndarray[np.float64_t,ndim=1] _cross_track_point(double  lon1, double  lat1, double  lon2, double  lat2, double  lon3, double  lat3):
    '''Get the closest point on great circle path to the 3rd point

    <lat1>, <lon1>: scalar float or nd-array, latitudes and longitudes in
                    degree, start point of the great circle.
    <lat2>, <lon2>: scalar float or nd-array, latitudes and longitudes in
                    degree, end point of the great circle.
    <lat3>, <lon3>: scalar float or nd-array, latitudes and longitudes in
                    degree, a point away from the great circle.

    Return <latp>, <lonp>: latitude and longitude of point P on the great
                           circle that connects P1, P2, and is closest
                           to point P3.
    '''
    cdef double x1,y1,z1,x2,y2,z2,x3,y3,z3,D,E,F,a,b,c,f,g,h,tt,xp,yp,zp,d1,d2

    x1,y1,z1=_spherical2Cart(lon1,lat1)
    x2,y2,z2=_spherical2Cart(lon2,lat2)
    x3,y3,z3=_spherical2Cart(lon3,lat3)

    D,E,F=np.cross([x1,y1,z1],[x2,y2,z2])

    a=E*z3-F*y3
    b=F*x3-D*z3
    c=D*y3-E*x3

    f=c*E-b*F
    g=a*F-c*D
    h=b*D-a*E

    tt=sqrt(f**2+g**2+h**2)
    xp=f/tt
    yp=g/tt
    zp=h/tt

    lon1, lat1 =_cart2Spherical(xp,yp,zp)
    lon2, lat2 =_cart2Spherical(-xp,-yp,-zp)
    d1=_great_circle_distance(lon1, lat1, lon3, lat3)
    d2=_great_circle_distance(lon2, lat2, lon3, lat3)

    if d1>d2:
        return np.array([lon2, lat2])
    else:
        return np.array([lon1, lat1])


def c_cross_track_point(double  lon1, double  lat1, double  lon2, double  lat2, double  lon3, double  lat3):
    '''Get the closest point on great circle path to the 3rd point

    <lat1>, <lon1>: scalar float or nd-array, latitudes and longitudes in
                    degree, start point of the great circle.
    <lat2>, <lon2>: scalar float or nd-array, latitudes and longitudes in
                    degree, end point of the great circle.
    <lat3>, <lon3>: scalar float or nd-array, latitudes and longitudes in
                    degree, a point away from the great circle.

    Return <latp>, <lonp>: latitude and longitude of point P on the great
                           circle that connects P1, P2, and is closest
                           to point P3.
    '''
    cdef double x1,y1,z1,x2,y2,z2,x3,y3,z3,D,E,F,a,b,c,f,g,h,tt,xp,yp,zp,d1,d2

    x1,y1,z1=_spherical2Cart(lon1,lat1)
    x2,y2,z2=_spherical2Cart(lon2,lat2)
    x3,y3,z3=_spherical2Cart(lon3,lat3)

    D,E,F=np.cross([x1,y1,z1],[x2,y2,z2])

    a=E*z3-F*y3
    b=F*x3-D*z3
    c=D*y3-E*x3

    f=c*E-b*F
    g=a*F-c*D
    h=b*D-a*E

    tt=sqrt(f**2+g**2+h**2)
    xp=f/tt
    yp=g/tt
    zp=h/tt

    lon1, lat1 =_cart2Spherical(xp,yp,zp)
    lon2, lat2 =_cart2Spherical(-xp,-yp,-zp)
    d1=_great_circle_distance(lon1, lat1, lon3, lat3)
    d2=_great_circle_distance(lon2, lat2, lon3, lat3)

    if d1>d2:
        return np.array([lon2, lat2])
    else:
        return np.array([lon1, lat1])


cdef double _cross_track_distance( double lon1, double lat1, double lon2, double lat2, double lon3, double lat3):
    """
    Usage
    -----
    Angular cross-track-distance of a point (lon3, lat3) from a great-circle path between (lon1, lat1) and (lon2, lat2)
    The sign of this distance tells which side of the path the third point is on.

    Parameters :

    param lat1: float, latitude of the first point
    param lon1: float, longitude of the first point
    param lat2: float, latitude of the second point
    param lon2: float, longitude of the second point
    param lat3: float, latitude of the third point
    param lon3: float, longitude of the third point

    Usage
    -----
    crt: float
         the (angular) c_cross_track_distance

    """
    cdef double d13,theta13,theta12,crt

    d13=_great_circle_distance(lon1,lat1,lon3,lat3) # distance from start point to third point
    theta13=_initial_bearing(lon1,lat1,lon3,lat3) # bearing from start point to third point
    theta12=_initial_bearing(lon1,lat1,lon2,lat2) # bearing from start point to end point

    crt= asin(sin(d13/R)*sin(theta13-theta12))*R

    return crt

def c_cross_track_distance( double lon1, double lat1, double lon2, double lat2, double lon3, double lat3):
    """
    Usage
    -----
    Angular cross-track-distance of a point (lon3, lat3) from a great-circle path between (lon1, lat1) and (lon2, lat2)
    The sign of this distance tells which side of the path the third point is on.

    Parameters :

    param lat1: float, latitude of the first point
    param lon1: float, longitude of the first point
    param lat2: float, latitude of the second point
    param lon2: float, longitude of the second point
    param lat3: float, latitude of the third point
    param lon3: float, longitude of the third point

    Usage
    -----
    crt: float
         the (angular) c_cross_track_distance

    """
    cdef double d13,theta13,theta12,crt


    d13=_great_circle_distance(lon1,lat1,lon3,lat3) # distance from start point to third point
    theta13=_initial_bearing(lon1,lat1,lon3,lat3) # bearing from start point to third point
    theta12=_initial_bearing(lon1,lat1,lon2,lat2) # bearing from start point to end point

    crt= asin(sin(d13/R)*sin(theta13-theta12))*R

    return crt

cdef double _along_track_distance(double crt,double lon1,double lat1,double lon3,double lat3):
    """
    Usage
    -----
    The along-track distance from the start point (lon1, lat1) to the closest point on the the path
    to the third point (lon3, lat3).

    Parameters
    ----------
    param lat1: float, latitude of the first point
    param lon1: float, longitude of the first point
    param lat3: float, latitude of the third point
    param lon3: float, longitude of the third point
    param crt : float, c_cross_track_distance

    Returns
    -------
    alt: float
         The along-track distance
    """

    cdef double d13,alt

    d13=_great_circle_distance(lon1,lat1,lon3,lat3)

    alt=acos(cos(d13/R)/cos(crt/R)) * R
    return alt

def c_along_track_distance(double crt,double lon1,double lat1,double lon3,double lat3):
    """
    Usage
    -----
    The along-track distance from the start point (lon1, lat1) to the closest point on the the path
    to the third point (lon3, lat3).

    Parameters
    ----------
    param lat1: float, latitude of the first point
    param lon1: float, longitude of the first point
    param lat3: float, latitude of the third point
    param lon3: float, longitude of the third point
    param crt : float, c_cross_track_distance

    Returns
    -------
    alt: float
         The along-track distance
    """

    cdef double d13,alt

    d13=_great_circle_distance(lon1,lat1,lon3,lat3)

    alt=acos(cos(d13/R)/cos(crt/R)) * R
    return alt


cdef double _point_to_path(double lon1,double lat1,double lon2,double lat2,double lon3,double lat3):
    """
    Usage
    -----
    The point-to-path distance between point (lon3, lat3) and path delimited by (lon1, lat1) and (lon2, lat2).
    The point-to-path distance is the cross_track distance between the great circle path if the projection of
    the third point lies on the path. If it is not on the path, return the minimum of the
    c_great_circle_distance between the first and the third or the second and the third point.

    Parameters
    ----------
    param lat1: float, latitude of the first point
    param lon1: float, longitude of the first point
    param lat2: float, latitude of the second point
    param lon2: float, longitude of the second point
    param lat3: float, latitude of the third point
    param lon3: float, longitude of the third point

    Returns
    -------

    ptp : float
          The point-to-path distance between point (lon3, lat3) and path delimited by (lon1, lat1) and (lon2, lat2)

    """
    cdef double crt,d1p,d2p,d12
    crt=_cross_track_distance(lon1,lat1,lon2,lat2,lon3,lat3)
    d1p=_along_track_distance(crt,lon1,lat1,lon3,lat3)
    d2p=_along_track_distance(crt,lon2,lat2,lon3,lat3)
    d12=_great_circle_distance(lon1,lat1,lon2,lat2)
    if (d1p > d12) or (d2p > d12):
        crt=fmin(_great_circle_distance(lon1,lat1,lon3,lat3),_great_circle_distance(lon2,lat2,lon3,lat3))
    else:
        crt=fabs(crt)
    return crt

def c_point_to_path(double lon1,double lat1,double lon2,double lat2,double lon3,double lat3):
    """
    Usage
    -----
    The point-to-path distance between point (lon3, lat3) and path delimited by (lon1, lat1) and (lon2, lat2).
    The point-to-path distance is the cross_track distance between the great circle path if the projection of
    the third point lies on the path. If it is not on the path, return the minimum of the
    c_great_circle_distance between the first and the third or the second and the third point.

    Parameters
    ----------
    param lat1: float, latitude of the first point
    param lon1: float, longitude of the first point
    param lat2: float, latitude of the second point
    param lon2: float, longitude of the second point
    param lat3: float, latitude of the third point
    param lon3: float, longitude of the third point

    Returns
    -------

    ptp : float
          The point-to-path distance between point (lon3, lat3) and path delimited by (lon1, lat1) and (lon2, lat2)

    """
    cdef double crt,d1p,d2p,d12
    crt=_cross_track_distance(lon1,lat1,lon2,lat2,lon3,lat3)
    d1p=_along_track_distance(crt,lon1,lat1,lon3,lat3)
    d2p=_along_track_distance(crt,lon2,lat2,lon3,lat3)
    d12=_great_circle_distance(lon1,lat1,lon2,lat2)
    if (d1p > d12) or (d2p > d12):
        crt=fmin(_great_circle_distance(lon1,lat1,lon3,lat3),_great_circle_distance(lon2,lat2,lon3,lat3))
    else:
        crt=fabs(crt)
    return crt



from .basic_euclidean import eucl_dist
from .basic_spherical import great_circle_distance


######################
# Euclidean Geometry #
######################

def e_lcss(t0, t1, eps):
    """
    Usage
    -----
    The Longuest-Common-Subsequence distance between trajectory t0 and t1.

    Parameters
    ----------
    param t0 : len(t0)x2 numpy_array
    param t1 : len(t1)x2 numpy_array
    eps : float

    Returns
    -------
    lcss : float
           The Longuest-Common-Subsequence distance between trajectory t0 and t1
    """
    n0 = len(t0)
    n1 = len(t1)
    # An (m+1) times (n+1) matrix
    C = [[0] * (n1 + 1) for _ in range(n0 + 1)]
    for i in range(1, n0 + 1):
        for j in range(1, n1 + 1):
            if eucl_dist(t0[i - 1], t1[j - 1]) < eps:
                C[i][j] = C[i - 1][j - 1] + 1
            else:
                C[i][j] = max(C[i][j - 1], C[i - 1][j])
    lcss = 1 - float(C[n0][n1]) / min([n0, n1])
    return lcss


######################
# Spherical Geometry #
######################

def s_lcss(t0, t1, eps):
    """
    Usage
    -----
    The Longuest-Common-Subsequence distance between trajectory t0 and t1.

    Parameters
    ----------
    param t0 : len(t0)x2 numpy_array
    param t1 : len(t1)x2 numpy_array
    eps : float

    Returns
    -------
    lcss : float
           The Longuest-Common-Subsequence distance between trajectory t0 and t1
    """
    n0 = len(t0)
    n1 = len(t1)
    # An (m+1) times (n+1) matrix
    C = [[0] * (n1 + 1) for _ in range(n0 + 1)]
    for i in range(1, n0 + 1):
        for j in range(1, n1 + 1):
            if great_circle_distance(t0[i - 1, 0], t0[i - 1, 1], t1[j - 1, 0], t1[j - 1, 1]) < eps:
                C[i][j] = C[i - 1][j - 1] + 1
            else:
                C[i][j] = max(C[i][j - 1], C[i - 1][j])
    lcss = 1 - float(C[n0][n1]) / min([n0, n1])
    return lcss



import pickle
import traj_dist.distance as tdist
from traj_dist.pydist.linecell import trajectory_set_grid
import timeit
import collections
import pandas as pd
import numpy as np

traj_list = pickle.load(open("/Users/bguillouet/These/trajectory_distance/data/benchmark_trajectories.pkl", "rb"))[:100]

time_dict = collections.defaultdict(dict)

for distance in ["sspd", "frechet", "discret_frechet", "hausdorff", "dtw", "lcss", "edr", "erp"]:
    t_euclidean = timeit.timeit(lambda: tdist.pdist(traj_list, metric=distance), number=1)

    if not (distance in ["frechet", "discret_frechet"]):
        t_spherical = timeit.timeit(
            lambda: tdist.pdist(traj_list, metric=distance, type_d="spherical"),  number=1)
    else:
        t_spherical = -1
    time_dict[distance] = {"Euclidean": t_euclidean, "Spherical": t_spherical}

t_cells_conversion_dic = collections.defaultdict(int)
for precision in [5, 6, 7]:
    cells_list_, _, _, _, _ = trajectory_set_grid(traj_list, precision=precision)
    cells_list = [np.array(x)[:, :2] for x in cells_list_]

    t_cells_conversion = timeit.timeit(lambda: trajectory_set_grid(traj_list, precision=7), number=1)
    t_cells_conversion_dic[precision] = t_cells_conversion

    t_euclidean = timeit.timeit(
        lambda: tdist.pdist(cells_list, metric="sowd_grid", type_d="euclidean", converted=True), number=1, )
    t_spherical = timeit.timeit(
        lambda: tdist.pdist(cells_list, metric="sowd_grid", type_d="spherical", converted=True), number=1, )

    time_dict["sowd_grid_%d"%precision] = {"Euclidean": t_euclidean, "Spherical": t_spherical}


df_cells_conversion = pd.Series(t_cells_conversion_dic)
df_cells_conversion.to_csv("/Users/bguillouet/These/trajectory_distance/data/benchmark_cells_conversion.csv")

df = pd.DataFrame(time_dict).transpose()
df.to_csv("/Users/bguillouet/These/trajectory_distance/data/benchmark.csv")



STUFF = "Hi"

from libc.math cimport fmax
from libc.math cimport fmin

cimport numpy as np

from basic_euclidean import c_point_to_trajectory
from basic_geographical import c_point_to_path

#############
# Euclidean #
#############

cdef double _e_directed_hausdorff ( np.ndarray[np.float64_t,ndim=2] t1,np.ndarray[np.float64_t,ndim=2] t2):
    """
    Usage
    -----
    directed hausdorff distance from trajectory t1 to trajectory t2.

    Parameters
    ----------
    param t1 :  len(t1)x2 numpy_array
    param t2 :  len(t2)x2 numpy_array

    Returns
    -------
    dh : float, directed hausdorff from trajectory t1 to trajectory t2
    """
    cdef double dh
    cdef int n1,i

    dh=0.0
    n1 = len(t1)
    for i from 0 <= i < n1:
        dh=fmax(c_point_to_trajectory(t1[i,0],t1[i,1],t2),dh)
    return dh

def c_e_directed_hausdorff ( np.ndarray[np.float64_t,ndim=2] t1,np.ndarray[np.float64_t,ndim=2] t2):
    """
    Usage
    -----
    directed hausdorff distance from trajectory t1 to trajectory t2.

    Parameters
    ----------
    param t1 :  len(t1)x2 numpy_array
    param t2 :  len(t2)x2 numpy_array

    Returns
    -------
    dh : float, directed hausdorff from trajectory t1 to trajectory t2
    """
    cdef double dh
    cdef int n1,i

    dh=0.0
    n1 = len(t1)
    for i from 0 <= i < n1:
        dh=fmax(c_point_to_trajectory(t1[i,0],t1[i,1],t2),dh)
    return dh


def c_e_hausdorff(np.ndarray[np.float64_t,ndim=2] t1,np.ndarray[np.float64_t,ndim=2] t2):
    """
    Usage
    -----
    hausdorff distance between trajectories t1 and t2.

    Parameters
    ----------
    param t1 :  len(t1)x2 numpy_array
    param t2 :  len(t2)x2 numpy_array

    Returns
    -------
    h : float, hausdorff from trajectories t1 and t2
    """
    cdef double h

    h=fmax(_e_directed_hausdorff(t1,t2),_e_directed_hausdorff(t2,t1))
    return h


################
# Geographical #
################


cdef double _g_directed_hausdorff ( np.ndarray[np.float64_t,ndim=2] t1,np.ndarray[np.float64_t,ndim=2] t2):
    """
    Usage
    -----
    directed hausdorff distance from trajectory t1 to trajectory t2.

    Parameters
    ----------
    param t1 :  len(t1)x2 numpy_array
    param t2 :  len(t2)x2 numpy_array

    Returns
    -------
    dh : float, directed hausdorff from trajectory t1 to trajectory t2
    """
    cdef double dh,dist_k0
    cdef int n0,n1,i,j
    cdef np.ndarray[np.float64_t,ndim=1] lats0,lons0,lats1,lons1

    n1 = len(t2)
    n0 = len(t1)
    lats0=t1[:,1]
    lons0=t1[:,0]
    lats1=t2[:,1]
    lons1=t2[:,0]

    dh=0
    for j from 0 <= j < n1  :
        dist_j0=9e100
        for i from 0 <= i < (n0-1):
            dist_j0=fmin(dist_j0,c_point_to_path(lons0[i],lats0[i],lons0[i+1],lats0[i+1],lons1[j],lats1[j]))
        dh=fmax(dh,dist_j0)
    return dh

def c_g_directed_hausdorff ( np.ndarray[np.float64_t,ndim=2] t1,np.ndarray[np.float64_t,ndim=2] t2):
    """
    Usage
    -----
    directed hausdorff distance from trajectory t1 to trajectory t2.

    Parameters
    ----------
    param t1 :  len(t1)x2 numpy_array
    param t2 :  len(t2)x2 numpy_array

    Returns
    -------
    dh : float, directed hausdorff from trajectory t1 to trajectory t2
    """
    cdef double dh,dist_k0
    cdef int n0,n1,i,j
    cdef np.ndarray[np.float64_t,ndim=1] lats0,lons0,lats1,lons1

    n1 = len(t2)
    n0 = len(t1)
    lats0=t1[:,1]
    lons0=t1[:,0]
    lats1=t2[:,1]
    lons1=t2[:,0]

    dh=0
    for j from 0 <= j < n1  :
        dist_j0=9e100
        for i from 0 <= i < (n0-1):
            dist_j0=fmin(dist_j0,c_point_to_path(lons0[i],lats0[i],lons0[i+1],lats0[i+1],lons1[j],lats1[j]))
        dh=fmax(dh,dist_j0)
    return dh

def c_g_hausdorff(np.ndarray[np.float64_t,ndim=2] t1,np.ndarray[np.float64_t,ndim=2] t2):
    """
    Usage
    -----
    hausdorff distance between trajectories t1 and t2.

    Parameters
    ----------
    param t1 :  len(t1)x2 numpy_array
    param t2 :  len(t2)x2 numpy_array

    Returns
    -------
    h : float, hausdorff from trajectories t1 and t2
    """
    cdef double h

    h=fmax(_g_directed_hausdorff(t1,t2),_g_directed_hausdorff(t2,t1))
    return h


import geohash2.geohash as geoh
import shapely.geometry as geos
import numpy as np


def linecell_lons_bigger_step(p1, p2, cell_start, lons_all, lats_all, lons_center_all, lats_center_all):
    reverse = False
    if p2[1] < p1[1]:
        tmp = p1
        p1 = p2
        p2 = tmp
        reverse = True

    lats_start_index = np.where(lats_all < p1[1])[0][-1]
    lats_end_index = np.where(lats_all > p2[1])[0][0]
    lats = lats_all[lats_start_index:lats_end_index + 1]

    if p1[0] < p2[0]:
        order = "croissant"
        idx_step = 1
        lons_start_index = np.where(lons_all < p1[0])[0][-1]
        lons_end_index = np.where(lons_all > p2[0])[0][0]
        lons = lons_all[lons_start_index:lons_end_index + 1]
    else:
        order = "decroissant"
        idx_step = -1
        lons_start_index = np.where(lons_all < p2[0])[0][-1]
        lons_end_index = np.where(lons_all > p1[0])[0][0]
        lons = lons_all[lons_start_index:lons_end_index + 1]
        lons = lons[::-1]
    line = geos.LineString([p1, p2])

    nlons = len(lons) - 2
    nlats = len(lats) - 2

    if not reverse:
        cell = [cell_start]
    else:
        if order == "croissant":
            cell = [[cell_start[0] - nlons, cell_start[1] - nlats]]
        else:
            cell = [[cell_start[0] + nlons, cell_start[1] - nlats]]

    lons_inter = []
    for l in lons[1:-1]:
        lons_shape = geos.LineString([[l, lats[0]], [l, lats[-1]]])
        lons_inter.append(lons_shape.intersection(line))

    idx_lat = 0
    for p_int in lons_inter:
        if p_int.y < lats[idx_lat + 1]:
            cell.append([cell[-1][0] + idx_step, cell[-1][1]])
        else:
            cell.append([cell[-1][0], cell[-1][1] + 1])
            cell.append([cell[-1][0] + idx_step, cell[-1][1]])
            idx_lat += 1
    if p2[1] > lats[idx_lat + 1]:
        cell.append([cell[-1][0], cell[-1][1] + 1])
    if reverse:
        cell.reverse()
    cells_coord = [[lons_center_all[x[0]], lats_center_all[x[1]]] for x in cell]
    return cell, cells_coord


def linecell_lats_bigger_step(p1, p2, cell_start, lons_all, lats_all, lons_center_all, lats_center_all):
    reverse = False
    if p2[0] < p1[0]:
        tmp = p1
        p1 = p2
        p2 = tmp
        reverse = True

    lons_start_index = np.where(lons_all < p1[0])[0][-1]
    lons_end_index = np.where(lons_all > p2[0])[0][0]
    lons = lons_all[lons_start_index:lons_end_index + 1]

    if p1[1] < p2[1]:
        order = "croissant"
        idx_step = 1
        lats_start_index = np.where(lats_all < p1[1])[0][-1]
        lats_end_index = np.where(lats_all > p2[1])[0][0]
        lats = lats_all[lats_start_index:lats_end_index + 1]
    else:
        order = "decroissant"
        idx_step = -1
        lats_start_index = np.where(lats_all < p2[1])[0][-1]
        lats_end_index = np.where(lats_all > p1[1])[0][0]
        lats = lats_all[lats_start_index:lats_end_index + 1]
        lats = lats[::-1]
    line = geos.LineString([p1, p2])

    nlons = len(lons) - 2
    nlats = len(lats) - 2

    if not reverse:
        cell = [cell_start]
    else:
        if order == "croissant":
            cell = [[cell_start[0] - nlons, cell_start[1] - nlats]]
        else:
            cell = [[cell_start[0] - nlons, cell_start[1] + nlats]]

    lats_inter = []
    for l in lats[1:-1]:
        lats_shape = geos.LineString([[lons[0], l], [lons[-1], l]])
        lats_inter.append(lats_shape.intersection(line))

    idx_lon = 0
    for p_int in lats_inter:
        if p_int.x < lons[idx_lon + 1]:
            cell.append([cell[-1][0], cell[-1][1] + idx_step])
        else:
            cell.append([cell[-1][0] + 1, cell[-1][1]])
            cell.append([cell[-1][0], cell[-1][1] + idx_step])
            idx_lon += 1
    if p2[0] > lons[idx_lon + 1]:
        cell.append([cell[-1][0] + 1, cell[-1][1]])
    if reverse:
        cell.reverse()
    cells_coord = [[lons_center_all[x[0]], lats_center_all[x[1]]] for x in cell]
    return cell, cells_coord


def get_extremum(traj):
    lons = traj[:, 0]
    lats = traj[:, 1]
    min_lon = min(lons)
    min_lat = min(lats)
    max_lon = max(lons)
    max_lat = max(lats)
    return min_lon, min_lat, max_lon, max_lat


def trajectory_set_grid(traj_set, precision, time=False):
    extremums = np.array(list(map(get_extremum, traj_set)))
    p_bottom_left = [min(extremums[:, 0]), min(extremums[:, 1])]
    p_top_right = [max(extremums[:, 2]), max(extremums[:, 3])]
    p_ble = geoh.encode(p_bottom_left[1], p_bottom_left[0], precision)
    p_tre = geoh.encode(p_top_right[1], p_top_right[0], precision)
    lat_ble, lon_ble, dlat, dlon = geoh.decode_exactly(p_ble)
    lat_tre, lon_tre, dlat, dlon = geoh.decode_exactly(p_tre)
    lats_all = np.arange(lat_ble - dlat, lat_tre + (3 * dlat), dlat * 2)
    lons_all = np.arange(lon_ble - dlon, lon_tre + 3 * dlon, dlon * 2)
    lats_center_all = np.arange(lat_ble, lat_tre + 2 * dlat, dlat * 2)
    lons_center_all = np.arange(lon_ble, lon_tre + 2 * dlon, dlon * 2)

    cells_traj = []
    for traj in traj_set:
        p_start = traj[0]
        cell_start_x = np.where(lons_all < p_start[0])[0][-1]
        cell_start_y = np.where(lats_all < p_start[1])[0][-1]
        cell_start = [cell_start_x, cell_start_y]

        cells = []

        for id_seg in range(len(traj) - 1):
            start = traj[id_seg]
            end = traj[id_seg + 1]
            if time:
                cell_start_time = start[2]
            if abs(start[0] - end[0]) / dlon > abs(start[1] - end[1]) / dlat:
                cell, cells_coord = linecell_lons_bigger_step(start, end, cell_start[:2], lons_all, lats_all,
                                                              lons_center_all,
                                                              lats_center_all)
            else:
                cell, cells_coord = linecell_lats_bigger_step(start, end, cell_start[:2], lons_all, lats_all,
                                                              lons_center_all,
                                                              lats_center_all)
            if time:
                if not cells:
                    cell_time = [cell[0] + [True, [cell_start_time]]]
                else:
                    if cell[0] == cells[-1][:2]:
                        cells[-1][3].append(cell_start_time)
                        cell_time = []
                    else:
                        cell_time = [cell[0] + [True, [cell_start_time]]]
                cell_time = cell_time + [x + [False, -1] for x in cell[1:-1]]
            else:
                if not cells:
                    cell_time = [cell[0] + [True]]
                else:
                    if cell[0] == cells[-1][:2]:
                        cell_time = []
                    else:
                        cell_time = [cell[0] + [True]]
                cell_time = cell_time + [x + [False] for x in cell[1:-1]]

            cells.extend(cell_time)
            cell_start = cell[-1]
        if time:
            cell_end_time = end[2]
            if cell_start == cells[-1][:2]:
                cells[-1][3].append(cell_end_time)
            else:
                cells.append(cell_start + [True, [cell_end_time]])
        else:
            if cell_start != cells[-1][:2]:
                cells.append(cell_start + [True])
        cells_traj.append(cells)
    # cells_traj_=map(np.array,cells_traj)
    return cells_traj, lons_all, lats_all, lons_center_all, lats_center_all


def trajectory_grid(traj_0, precision):
    cells_list, lons_all, lats_all, lons_center_all, lats_center_all = trajectory_set_grid([traj_0], precision)
    return cells_list[0], lons_all, lats_all, lons_center_all, lats_center_all



STUFF = "Hi"

from libc.math cimport fmin,fabs

cimport numpy as np
import numpy as np
from numpy.math cimport INFINITY

from basic_euclidean import c_eucl_dist
from basic_geographical import c_great_circle_distance

###############
#### ERP ######
###############

def c_e_erp(np.ndarray[np.float64_t,ndim=2] t0, np.ndarray[np.float64_t,ndim=2] t1, np.ndarray[np.float64_t,ndim=1] g):
    """
    Usage
    -----
    The Edit distance with Real Penalty between trajectory t0 and t1.

    Parameters
    ----------
    param t0 : len(t0)x2 numpy_array
    param t1 : len(t1)x2 numpy_array

    Returns
    -------
    dtw : float
          The Dynamic-Time Warping distance between trajectory t0 and t1
    """
    cdef int n0,n1,i,j
    cdef np.ndarray[np.float64_t,ndim=2] C
    cdef double x0,y0,x1,y1,dtw,gx,gy,edgei,edgej

    gx=g[0]
    gy=g[1]

    n0 = len(t0)+1
    n1 = len(t1)+1
    C=np.zeros((n0,n1))
    edgei=0
    for i from 1 <= i < n0:
        x0=t0[i-1,0]
        y0=t0[i-1,1]
        edgei += fabs(c_eucl_dist(x0,y0,gx,gy))
    C[1:,0]=edgei

    edgej=0
    for j from 1 <= j < n1:
        x1=t1[j-1,0]
        y1=t1[j-1,1]
        edgej += fabs(c_eucl_dist(x1,y1,gx,gy))
    C[0,1:]=edgej

    for i from 1 <= i < n0:
        for j from 1 <= j < n1:
            x0=t0[i-1,0]
            y0=t0[i-1,1]
            x1=t1[j-1,0]
            y1=t1[j-1,1]
            derp0 = C[i-1,j] + c_eucl_dist(x0,y0,gx,gy)
            derp1 = C[i,j-1] + c_eucl_dist(gx,gy,x1,y1)
            derp01 = C[i-1,j-1] + c_eucl_dist(x0,y0,x1,y1)
            C[i,j] = fmin(derp0,fmin(derp1,derp01))
    erp = C[n0-1,n1-1]
    return erp


def c_g_erp(np.ndarray[np.float64_t,ndim=2] t0, np.ndarray[np.float64_t,ndim=2] t1, np.ndarray[np.float64_t,ndim=1] g):
    """
    Usage
    -----
    The Edit distance with Real Penalty between trajectory t0 and t1.

    Parameters
    ----------
    param t0 : len(t0)x2 numpy_array
    param t1 : len(t1)x2 numpy_array

    Returns
    -------
    dtw : float
          The Dynamic-Time Warping distance between trajectory t0 and t1
    """
    cdef int n0,n1,i,j
    cdef np.ndarray[np.float64_t,ndim=2] C
    cdef double x0,y0,x1,y1,dtw,gx,gy,edgei,edgej

    gx=g[0]
    gy=g[1]

    n0 = len(t0)+1
    n1 = len(t1)+1
    C=np.zeros((n0,n1))

    edgei=0.0
    for i from 1 <= i < n0:
        x0=t0[i-1,0]
        y0=t0[i-1,1]
        edgei += fabs(c_great_circle_distance(x0,y0,gx,gy))
    C[1:,0]=edgei

    edgej=0.0
    for j from 1 <= j < n1:
        x1=t1[j-1,0]
        y1=t1[j-1,1]
        edgej += fabs(c_great_circle_distance(x1,y1,gx,gy))
    C[0,1:]=edgej
    for i from 1 <= i < n0:
        for j from 1 <= j < n1:
            x0=t0[i-1,0]
            y0=t0[i-1,1]
            x1=t1[j-1,0]
            y1=t1[j-1,1]
            derp0 = C[i-1,j] + c_great_circle_distance(x0,y0,gx,gy)
            derp1 = C[i,j-1] + c_great_circle_distance(gx,gy,x1,y1)
            derp01 = C[i-1,j-1] + c_great_circle_distance(x0,y0,x1,y1)
            C[i,j] = fmin(derp0,fmin(derp1,derp01))
    erp = C[n0-1,n1-1]
    return erp



import numpy as np
from . import linecell as linec


def owd_grid_brut(traj_cell_1, traj_cell_2):
    """
    Usage
    -----
    The owd-distance of trajectory t2 from trajectory t1

    Parameters
    ----------
    param traj_cell_1 :  len(t1)x2 numpy_array
    param traj_cell_2 :  len(t2)x2 numpy_array

    Returns
    -------
    owd : float
           owd-distance of trajectory t2 from trajectory t1
    """
    D = 0
    n = len(traj_cell_1)
    for p1 in traj_cell_1:
        d = [np.linalg.norm(p1 - x) for x in traj_cell_2]
        D += min(d)
    owd = D / n
    return owd


def find_first_min_points(pt, n):
    """
    Usage
    -----
    Return the index of the min-point in the vector pt of size n.

    Parameters
    ----------
    param pt :  len(t1)x1 numpy_array
    param n :  int

    Returns
    -------
    min_point_index  : nbumber of min points x1 numpy_array

    """
    if n == 1:
        min_points = [True]
    else:
        min_points = [pt[0] < pt[1]]
        for i in range(1, n - 1):
            m_p = pt[i] < pt[i + 1] and pt[i] < pt[i - 1]
            min_points.append(m_p)
        min_points.append(pt[n - 1] < pt[n - 2])
    min_point_index = np.where(np.array(min_points))[0]
    return min_point_index


def owd_grid(traj_cell_1, traj_cell_2):
    """
    Usage
    -----
    The owd-distance of trajectory t2 from trajectory t1

    Parameters
    ----------
    param traj_cell_1 :  len(t1)x2 numpy_array
    param traj_cell_2 :  len(t2)x2 numpy_array

    Returns
    -------
    owd : float
           owd-distance of trajectory t2 from trajectory t1
    """
    n1 = len(traj_cell_1)
    n2 = len(traj_cell_2)

    p = traj_cell_1[0]
    p_t2 = [np.linalg.norm(p - x) for x in traj_cell_2]
    S_old = find_first_min_points(p_t2, n2)
    D = min(p_t2)
    for i in range(1, n1):
        p_prec = p
        p = traj_cell_1[i]
        S = []
        d = []
        n_S_old = len(S_old)
        for j in range(n_S_old):
            ig = S_old[j]
            pg = traj_cell_2[ig]
            if (p_prec[1] == p[1]) and (pg[0] != p_prec[0]) or (p_prec[0] == p[0]) and (pg[1] != p_prec[1]):
                S.append(ig)
                d.append(np.linalg.norm(p - pg))
            else:
                if j == 0:
                    if n_S_old == 1:
                        ranges = list(range(0, n2))
                    else:
                        ranges = list(range(0, S_old[j + 1]))
                elif j == n_S_old - 1:
                    ranges = list(range(S_old[j - 1], n2))
                else:
                    ranges = list(range(S_old[j - 1] + 1, S_old[j + 1]))
                for igp in ranges:
                    pgp = traj_cell_2[igp]
                    if (p_prec[1] == p[1] and pgp[0] == p[0]) or (p_prec[0] == p[0] and pgp[1] == p[1]) or igp == ig:
                        dist_back = np.linalg.norm(traj_cell_2[igp - 1] - p) if igp != 0 else np.inf
                        dist_forw = np.linalg.norm(traj_cell_2[igp + 1] - p) if igp != n2 - 1 else np.inf
                        dist = np.linalg.norm(pgp - p)
                        if dist < dist_back and dist < dist_forw:
                            if not (igp in S):
                                S.append(igp)
                                d.append(dist)
        S_old = S
        # p_t2 = map(lambda x: np.linalg.norm(traj_cell_1[i] - x), traj_cell_2)
        # print(np.all(map(lambda x,y:x==y,S_old,find_first_min_points(p_t2, n2))))
        D += min(d)
    return D / n1


def sowd_grid(traj_cell_1, traj_cell_2):
    sowd_dist = owd_grid(traj_cell_1, traj_cell_2) + owd_grid(traj_cell_2, traj_cell_1)
    return sowd_dist / 2


def sowd_grid_brut(traj_cell_1, traj_cell_2):
    sowd_brut_dist = owd_grid_brut(traj_cell_1, traj_cell_2) + owd_grid_brut(traj_cell_2, traj_cell_1)
    return sowd_brut_dist / 2


def sowd(traj_1, traj_2, precision=7, converted=False):
    if converted:
        d = sowd_grid(traj_1, traj_2)
    else:
        cells_list, _, _ = linec.trajectory_set_grid([traj_1, traj_2], precision)
        d = sowd_grid(cells_list[0], cells_list[1])
    return d


def sowd_brut(traj_1, traj_2, precision=7, converted=False):
    if converted:
        d = sowd_grid_brut(traj_1, traj_2)
    else:
        cells_list, _, _ = linec.trajectory_set_grid([traj_1, traj_2], precision)
        d = sowd_grid_brut(cells_list[0], cells_list[1])
    return d



import numpy as np
from .basic_euclidean import eucl_dist
from .basic_spherical import great_circle_distance


######################
# Euclidean Geometry #
######################

def e_dtw(t0, t1):
    """
    Usage
    -----
    The Dynamic-Time Warping distance between trajectory t0 and t1.

    Parameters
    ----------
    param t0 : len(t0)x2 numpy_array
    param t1 : len(t1)x2 numpy_array

    Returns
    -------
    dtw : float
          The Dynamic-Time Warping distance between trajectory t0 and t1
    """

    n0 = len(t0)
    n1 = len(t1)
    C = np.zeros((n0 + 1, n1 + 1))
    C[1:, 0] = float('inf')
    C[0, 1:] = float('inf')
    for i in np.arange(n0) + 1:
        for j in np.arange(n1) + 1:
            C[i, j] = eucl_dist(t0[i - 1], t1[j - 1]) + min(C[i, j - 1], C[i - 1, j - 1], C[i - 1, j])
    dtw = C[n0, n1]
    return dtw


######################
# Spherical Geometry #
######################

def s_dtw(t0, t1):
    """
    Usage
    -----
    The Dynamic-Time Warping distance between trajectory t0 and t1.

    Parameters
    ----------
    param t0 : len(t0)x2 numpy_array
    param t1 : len(t1)x2 numpy_array

    Returns
    -------
    dtw : float
          The Dynamic-Time Warping distance between trajectory t0 and t1
    """
    n0 = len(t0)
    n1 = len(t1)
    C = np.zeros((n0 + 1, n1 + 1))
    C[1:, 0] = float('inf')
    C[0, 1:] = float('inf')
    for i in np.arange(n0) + 1:
        for j in np.arange(n1) + 1:
            C[i, j] = great_circle_distance(t0[i - 1][0], t0[i - 1][1], t1[j - 1][0], t1[j - 1][1]) + \
                      min(C[i, j - 1], C[i - 1, j - 1], C[i - 1, j])
    dtw = C[n0, n1]
    return dtw



import numpy as np
from .basic_euclidean import eucl_dist, eucl_dist_traj
from .basic_spherical import great_circle_distance


######################
# Euclidean Geometry #
######################

def e_erp(t0, t1, g):
    """
    Usage
    -----
    The Edit distance with Real Penalty between trajectory t0 and t1.

    Parameters
    ----------
    param t0 : len(t0)x2 numpy_array
    param t1 : len(t1)x2 numpy_array

    Returns
    -------
    dtw : float
          The Dynamic-Time Warping distance between trajectory t0 and t1
    """

    n0 = len(t0)
    n1 = len(t1)
    C = np.zeros((n0 + 1, n1 + 1))

    gt0_dist = [abs(eucl_dist(g, x)) for x in t0]
    gt1_dist = [abs(eucl_dist(g, x)) for x in t1]
    mdist = eucl_dist_traj(t0, t1)

    C[1:, 0] = sum(gt0_dist)
    C[0, 1:] = sum(gt1_dist)
    for i in np.arange(n0) + 1:
        for j in np.arange(n1) + 1:
            derp0 = C[i - 1, j] + gt0_dist[i-1]
            derp1 = C[i, j - 1] + gt1_dist[j-1]
            derp01 = C[i - 1, j - 1] + mdist[i-1, j-1]
            C[i, j] = min(derp0, derp1, derp01)
    erp = C[n0, n1]
    return erp


######################
# Spherical Geometry #
######################

def s_erp(t0, t1, g):
    """
    Usage
    -----
    The Edit distance with Real Penalty between trajectory t0 and t1.

    Parameters
    ----------
    param t0 : len(t0)x2 numpy_array
    param t1 : len(t1)x2 numpy_array

    Returns
    -------
    dtw : float
          The Dynamic-Time Warping distance between trajectory t0 and t1
    """
    n0 = len(t0)
    n1 = len(t1)
    C = np.zeros((n0 + 1, n1 + 1))

    C[1:, 0] = sum([abs(great_circle_distance(g[0], g[1], x[0], x[1])) for x in t0])
    C[0, 1:] = sum([abs(great_circle_distance(g[0], g[1], y[0], y[1])) for y in t1])
    for i in np.arange(n0) + 1:
        for j in np.arange(n1) + 1:
            derp0 = C[i - 1, j] + great_circle_distance(t0[i - 1][0], t0[i - 1][1], g[0], g[1])
            derp1 = C[i, j - 1] + great_circle_distance(g[0], g[1], t1[j - 1][0], t1[j - 1][1])
            derp01 = C[i - 1, j - 1] + great_circle_distance(t0[i - 1][0], t0[i - 1][1], t1[j - 1][0], t1[j - 1][1])
            C[i, j] = min(derp0, derp1, derp01)
    erp = C[n0, n1]
    return erp



STUFF = "Hi"

from libc.math cimport fmin
from libc.math cimport fmax


cimport numpy as np
import numpy as np
from numpy.math cimport INFINITY

from basic_euclidean import c_eucl_dist
from basic_geographical import c_great_circle_distance

###########################
#### DISCRET FRECHET ######
###########################

def c_discret_frechet(np.ndarray[np.float64_t,ndim=2] t0, np.ndarray[np.float64_t,ndim=2] t1):
    """
    Usage
    -----
    The Dynamic-Time Warping distance between trajectory t0 and t1.

    Parameters
    ----------
    param t0 : len(t0)x2 numpy_array
    param t1 : len(t1)x2 numpy_array

    Returns
    -------
    dtw : float
          The Dynamic-Time Warping distance between trajectory t0 and t1
    """
    cdef int n0,n1,i,j
    cdef np.ndarray[np.float64_t,ndim=2] C
    cdef double x0,y0,x1,y1,dtw

    n0 = len(t0)+1
    n1 = len(t1)+1
    C=np.zeros((n0,n1))

    C[1:,0]=INFINITY
    C[0,1:]=INFINITY
    for i from 1 <= i < n0:
        for j from 1 <= j < n1:
            x0=t0[i-1,0]
            y0=t0[i-1,1]
            x1=t1[j-1,0]
            y1=t1[j-1,1]
            C[i,j]=fmax(c_eucl_dist(x0,y0,x1,y1),fmin(fmin(C[i,j-1],C[i-1,j-1]),C[i-1,j]))
    dfr = C[n0-1,n1-1]
    return dfr




# Computing the Frechet distance between two polygonal curves.
# International Journal of Computational Geometry & Applications.
# Vol 5, Nos 1&2 (1995) 75-91

from .basic_euclidean import eucl_dist, point_to_seg, circle_line_intersection, eucl_dist_traj


def free_line(p, eps, s, dps1, dps2, ds):
    """
    Usage
    -----
    Return the free space in the segment s, from point p.
    This free space is the set of all point in s whose distance from p is at most eps.
    Since s is a segment, the free space is also a segment.
    We return a 1x2 array whit the fraction of the segment s which are in the free space.
    If no part of s are in the free space, return [-1,-1]

    Parameters
    ----------
    param p : 1x2 numpy_array, centre of the circle
    param eps : float, radius of the circle
    param s : 2x2 numpy_array, line

    Returns
    -------
    lf : 1x2 numpy_array
         fraction of segment which is in the free space (i.e [0.3,0.7], [0.45,1], ...)
         If no part of s are in the free space, return [-1,-1]
    """
    px = p[0]
    py = p[1]
    s1x = s[0, 0]
    s1y = s[0, 1]
    s2x = s[1, 0]
    s2y = s[1, 1]
    if s1x == s2x and s1y == s2y:
        if eucl_dist(p, s[0]) > eps:
            lf = [-1, -1]
        else:
            lf = [0, 1]
    else:
        if point_to_seg(p, s[0], s[1], dps1, dps2, ds) > eps:
            # print("No Intersection")
            lf = [-1, -1]
        else:
            segl = eucl_dist(s[0], s[1])
            segl2 = segl * segl
            intersect = circle_line_intersection(px, py, s1x, s1y, s2x, s2y, eps)
            if intersect[0][0] != intersect[1][0] or intersect[0][1] != intersect[1][1]:
                i1x = intersect[0, 0]
                i1y = intersect[0, 1]
                u1 = (((i1x - s1x) * (s2x - s1x)) + ((i1y - s1y) * (s2y - s1y))) / segl2

                i2x = intersect[1, 0]
                i2y = intersect[1, 1]
                u2 = (((i2x - s1x) * (s2x - s1x)) + ((i2y - s1y) * (s2y - s1y))) / segl2
                ordered_point = sorted((0, 1, u1, u2))
                lf = ordered_point[1:3]
            else:
                if px == s1x and py == s1y:
                    lf = [0, 0]
                elif px == s2x and py == s2y:
                    lf = [1, 1]
                else:
                    i1x = intersect[0][0]
                    i1y = intersect[0][1]
                    u1 = (((i1x - s1x) * (s2x - s1x)) + ((i1y - s1y) * (s2y - s1y))) / segl2
                    if 0 <= u1 <= 1:
                        lf = [u1, u1]
                    else:
                        lf = [-1, -1]
    return lf


def LF_BF(P, Q, p, q, eps, mdist, P_dist, Q_dist):
    """
    Usage
    -----
    Compute all the free space on the boundary of cells in the diagram for polygonal chains P and Q and the given eps
    LF[(i,j)] is the free space of segment [Pi,Pi+1] from point  Qj
    BF[(i,j)] is the free space of segment [Qj,Qj+1] from point Pj

    Parameters
    ----------
    param P : px2 numpy_array, Trajectory P
    param Q : qx2 numpy_array, Trajectory Q
    param p : float, number of points in Trajectory P
    param q : float, number of points in Trajectory Q
    param eps : float, reachability distance
    mdist : p x q numpy array, pairwise distance between points of trajectories t1 and t2
    param P_dist:  p x 1 numpy_array,  distances between consecutive points in P
    param Q_dist:  q x 1 numpy_array,  distances between consecutive points in Q

    Returns
    -------
    LF : dict, free spaces of segments of P from points of Q
    BF : dict, free spaces of segments of Q from points of P
    """
    LF = {}
    for j in range(q):
        for i in range(p - 1):
            LF.update({(i, j): free_line(Q[j], eps, P[i:i + 2], mdist[i, j], mdist[i + 1, j], P_dist[i])})
    BF = {}
    for j in range(q - 1):
        for i in range(p):
            BF.update({(i, j): free_line(P[i], eps, Q[j:j + 2], mdist[i, j], mdist[i, j + 1], Q_dist[j])})
    return LF, BF


def LR_BR(LF, BF, p, q):
    """
    Usage
    -----
    Compute all the free space,that are reachable from the origin (P[0,0],Q[0,0]) on the boundary of cells
    in the diagram for polygonal chains P and Q and the given free spaces LR and BR

    LR[(i,j)] is the free space, reachable from the origin, of segment [Pi,Pi+1] from point  Qj
    BR[(i,j)] is the free space, reachable from the origin, of segment [Qj,Qj+1] from point Pj

    Parameters
    ----------
    LF : dict, free spaces of segments of P from points of Q
    BF : dict, free spaces of segments of Q from points of P
    param p : float, number of points in Trajectory P
    param q : float, number of points in Trajectory Q

    Returns
    -------
    rep : bool, return true if frechet distance is inf to eps
    LR : dict, is the free space, reachable from the origin, of segments of P from points of Q
    BR : dict, is the free space, reachable from the origin, of segments of Q from points of P
    """
    if not (LF[(0, 0)][0] <= 0 and BF[(0, 0)][0] <= 0 and LF[(p - 2, q - 1)][1] >= 1 and BF[(p - 1, q - 2)][1] >= 1):
        rep = False
        BR = {}
        LR = {}
    else:
        LR = {(0, 0): True}
        BR = {(0, 0): True}
        for i in range(1, p - 1):
            if LF[(i, 0)] != [-1, -1] and LF[(i - 1, 0)] == [0, 1]:
                LR[(i, 0)] = True
            else:
                LR[(i, 0)] = False
        for j in range(1, q - 1):
            if BF[(0, j)] != [-1, -1] and BF[(0, j - 1)] == [0, 1]:
                BR[(0, j)] = True
            else:
                BR[(0, j)] = False
        for i in range(p - 1):
            for j in range(q - 1):
                if LR[(i, j)] or BR[(i, j)]:
                    if LF[(i, j + 1)] != [-1, -1]:
                        LR[(i, j + 1)] = True
                    else:
                        LR[(i, j + 1)] = False
                    if BF[(i + 1, j)] != [-1, -1]:
                        BR[(i + 1, j)] = True
                    else:
                        BR[(i + 1, j)] = False
                else:
                    LR[(i, j + 1)] = False
                    BR[(i + 1, j)] = False
        rep = BR[(p - 2, q - 2)] or LR[(p - 2, q - 2)]
    return rep, LR, BR


def decision_problem(P, Q, p, q, eps, mdist, P_dist, Q_dist):
    """
    Usage
    -----
    Test is the frechet distance between trajectories P and Q are inferior to eps

    Parameters
    ----------
    param P : px2 numpy_array, Trajectory P
    param Q : qx2 numpy_array, Trajectory Q
    param p : float, number of points in Trajectory P
    param q : float, number of points in Trajectory Q
    param eps : float, reachability distance
    mdist : p x q numpy array, pairwise distance between points of trajectories t1 and t2
    param P_dist:  p x 1 numpy_array,  distances between consecutive points in P
    param Q_dist:  q x 1 numpy_array,  distances between consecutive points in Q

    Returns
    -------
    rep : bool, return true if frechet distance is inf to eps
    """
    LF, BF = LF_BF(P, Q, p, q, eps, mdist, P_dist, Q_dist)
    rep, _, _ = LR_BR(LF, BF, p, q)
    return rep


def compute_critical_values(P, Q, p, q, mdist, P_dist, Q_dist):
    """
    Usage
    -----
    Compute all the critical values between trajectories P and Q

    Parameters
    ----------
    param P : px2 numpy_array, Trajectory P
    param Q : qx2 numpy_array, Trajectory Q
    param p : int, number of points in Trajectory P
    param q : int, number of points in Trajectory Q
    mdist : p x q numpy array, pairwise distance between points of trajectories t1 and t2
    param P_dist:  p x 1 numpy_array,  distances between consecutive points in P
    param Q_dist:  q x 1 numpy_array,  distances between consecutive points in Q

    Returns
    -------
    cc : list, all critical values between trajectories P and Q
    """
    origin = eucl_dist(P[0], Q[0])
    end = eucl_dist(P[-1], Q[-1])
    end_point = max(origin, end)
    cc = set([end_point])
    for i in range(p - 1):
        for j in range(q - 1):
            Lij = point_to_seg(Q[j], P[i], P[i + 1], mdist[i, j], mdist[i + 1, j], P_dist[i])
            if Lij > end_point:
                cc.add(Lij)
            Bij = point_to_seg(P[i], Q[j], Q[j + 1], mdist[i, j], mdist[i, j + 1], Q_dist[j])
            if Bij > end_point:
                cc.add(Bij)
    return sorted(list(cc))


def frechet(P, Q):
    """
    Usage
    -----
    Compute the frechet distance between trajectories P and Q

    Parameters
    ----------
    param P : px2 numpy_array, Trajectory P
    param Q : qx2 numpy_array, Trajectory Q

    Returns
    -------
    frech : float, the frechet distance between trajectories P and Q
    """
    p = len(P)
    q = len(Q)

    mdist = eucl_dist_traj(P, Q)
    P_dist = [eucl_dist(P[ip], P[ip + 1]) for ip in range(p - 1)]
    Q_dist = [eucl_dist(Q[iq], Q[iq + 1]) for iq in range(q - 1)]

    cc = compute_critical_values(P, Q, p, q, mdist, P_dist, Q_dist)
    eps = cc[0]
    while (len(cc) != 1):
        m_i = len(cc) / 2 - 1
        eps = cc[m_i]
        rep = decision_problem(P, Q, p, q, eps, mdist, P_dist, Q_dist)
        if rep:
            cc = cc[:m_i + 1]
        else:
            cc = cc[m_i + 1:]
    frech = eps
    return frech



import numpy as np
from .basic_euclidean import eucl_dist


def discret_frechet(t0, t1):
    """
    Usage
    -----
    Compute the discret frechet distance between trajectories P and Q

    Parameters
    ----------
    param t0 : px2 numpy_array, Trajectory t0
    param t1 : qx2 numpy_array, Trajectory t1

    Returns
    -------
    frech : float, the discret frechet distance between trajectories t0 and t1
    """
    n0 = len(t0)
    n1 = len(t1)
    C = np.zeros((n0 + 1, n1 + 1))
    C[1:, 0] = float('inf')
    C[0, 1:] = float('inf')
    for i in np.arange(n0) + 1:
        for j in np.arange(n1) + 1:
            C[i, j] = max(eucl_dist(t0[i - 1], t1[j - 1]), min(C[i, j - 1], C[i - 1, j - 1], C[i - 1, j]))
    dtw = C[n0, n1]
    return dtw



STUFF = "Hi"

from libc.math cimport fmax

cimport numpy as np
import numpy as np
from cpython cimport bool


from basic_euclidean import c_circle_line_intersection,c_point_to_seg,c_eucl_dist


cdef np.ndarray[np.float64_t,ndim=1] _free_line(double px, double py, double eps, double s1x, double s1y, double
s2x, double s2y):
    """
    Usage
    -----
    Return the free space in the segment s, from point p.
    This free space is the set of all point in s whose distance from p is at most eps.
    Since s is a segment, the free space is also a segment.
    We return a 1x2 array whit the fraction of the segment s which are in the free space.
    If no part of s are in the free space, return [-1,-1]

    Parameters
    ----------
    param px : float, centre's abscissa of the circle
    param py : float, centre's ordinate of the circle
    param eps : float, radius of the circle
    param s1x : abscissa of the first end point of the segment
    param s1y : ordinate of the first end point of the segment
    param s2x : abscissa of the second end point of the segment
    param s2y : ordinate of the second end point of the segment

    Returns
    -------
    lf : 1x2 numpy_array
         fraction of segment which is in the free space (i.e [0.3,0.7], [0.45,1], ...)
         If no part of s are in the free space, return [-1,-1]
    """
    cdef np.ndarray[np.float64_t,ndim=2] intersect
    cdef np.ndarray[np.float64_t,ndim=1] lf,ordered_point
    cdef double i1x,i1y,u1,i2x,i2y,u2,segl,segl2
    cdef bool pts_sup_eps,i1x_dif_i2x,i1y_dif_i2y,u1_sup_0,u2_inf_1

    pts_sup_eps = c_point_to_seg(px,py,s1x,s1y,s2x,s2y)>eps
    if s1x == s2x and s1y==s2y:
        if c_eucl_dist(px,py,s1x,s1y) > eps:
            lf = np.array([-1.0, -1.0])
        else:
            lf = np.array([0.0,1.0])
    else:
        if pts_sup_eps:
            #print("No Intersection")
            lf=np.array([-1.0,-1.0])
        else :
            segl=c_eucl_dist(s1x,s1y,s2x,s2y)
            segl2=segl*segl
            intersect = c_circle_line_intersection(px,py,s1x,s1y,s2x,s2y,eps)
            i1x = intersect[0,0]
            i2x = intersect[1,0]
            i1y = intersect[0,1]
            i2y = intersect[1,1]
            i1x_dif_i2x = i1x!=i2x
            i1y_dif_i2y = i1y!=i2y
            if i1x_dif_i2x or i1y_dif_i2y:
                u1 = (((i1x - s1x) * (s2x - s1x)) + ((i1y - s1y) * (s2y - s1y)))/segl2
                u2 = (((i2x - s1x) * (s2x - s1x)) + ((i2y - s1y) * (s2y - s1y)))/segl2
                ordered_point=np.array(sorted((0,1,u1,u2)))
                lf= ordered_point[1:3]
            else :
                if px == s1x and py==s1y:
                    lf = np.array([0.0, 0.0])
                elif px == s2x and py==s2y:
                    lf = np.array([1.0, 1.0])
                else:
                    u1 = (((i1x - s1x) * (s2x - s1x)) + ((i1y - s1y) * (s2y - s1y)))/segl2
                    u1_sup_0 = u1 >=0
                    u2_inf_1 = u1 <=1
                    if u1_sup_0 and u2_inf_1:
                        lf=np.array([u1,u1])
                    else:
                        lf=np.array([-1.0,-1.0])
    return lf


def c_free_line(double px, double py, double eps, double s1x, double s1y, double s2x, double s2y):
    """
    Usage
    -----
    Return the free space in the segment s, from point p.
    This free space is the set of all point in s whose distance from p is at most eps.
    Since s is a segment, the free space is also a segment.
    We return a 1x2 array whit the fraction of the segment s which are in the free space.
    If no part of s are in the free space, return [-1,-1]

    Parameters
    ----------
    param px : float, centre's abscissa of the circle
    param py : float, centre's ordinate of the circle
    param eps : float, radius of the circle
    param s1x : abscissa of the first end point of the segment
    param s1y : ordinate of the first end point of the segment
    param s2x : abscissa of the second end point of the segment
    param s2y : ordinate of the second end point of the segment

    Returns
    -------
    lf : 1x2 numpy_array
         fraction of segment which is in the free space (i.e [0.3,0.7], [0.45,1], ...)
         If no part of s are in the free space, return [-1,-1]
    """
    cdef np.ndarray[np.float64_t,ndim=2] intersect
    cdef np.ndarray[np.float64_t,ndim=1] lf,ordered_point
    cdef double i1x,i1y,u1,i2x,i2y,u2,segl,segl2
    cdef bool pts_sup_eps,i1x_dif_i2x,i1y_dif_i2y,u1_sup_0,u2_inf_1

    pts_sup_eps = c_point_to_seg(px,py,s1x,s1y,s2x,s2y)>eps

    if s1x == s2x and s1y==s2y:
        if c_eucl_dist(px,py,s1x,s1y) > eps:
            lf = np.array([-1.0, -1.0])
        else:
            lf = np.array([0.0,1.0])
    else:
        if pts_sup_eps:
            #print("No Intersection")
            lf=np.array([-1.0,-1.0])
        else :
            segl=c_eucl_dist(s1x,s1y,s2x,s2y)
            segl2=segl*segl
            intersect = c_circle_line_intersection(px,py,s1x,s1y,s2x,s2y,eps)
            i1x = intersect[0,0]
            i2x = intersect[1,0]
            i1y = intersect[0,1]
            i2y = intersect[1,1]
            i1x_dif_i2x = i1x!=i2x
            i1y_dif_i2y = i1y!=i2y
            if i1x_dif_i2x or i1y_dif_i2y:
                u1 = (((i1x - s1x) * (s2x - s1x)) + ((i1y - s1y) * (s2y - s1y)))/segl2
                u2 = (((i2x - s1x) * (s2x - s1x)) + ((i2y - s1y) * (s2y - s1y)))/segl2
                ordered_point=np.array(sorted((0,1,u1,u2)))
                lf= ordered_point[1:3]
            else :
                if px == s1x and py==s1y:
                    lf = np.array([0.0, 0.0])
                elif px == s2x and py==s2y:
                    lf = np.array([1.0, 1.0])
                else:
                    u1 = (((i1x - s1x) * (s2x - s1x)) + ((i1y - s1y) * (s2y - s1y)))/segl2
                    u1_sup_0 = u1 >=0
                    u2_inf_1 = u1 <=1
                    if u1_sup_0 and u2_inf_1:
                        lf=np.array([u1,u1])
                    else:
                        lf=np.array([-1.0,-1.0])
    return lf

cdef _compute_LF_BF(np.ndarray[np.float64_t,ndim=2] P, np.ndarray[np.float64_t,ndim=2] Q, int p, int q,
double eps):
    """
    Usage
    -----
    Compute all the free space on the boundary of cells in the diagram for polygonal chains P and Q and the given eps
    LF[(i,j)] is the free space of segments [Pi,Pi+1] from point  Qj
    BF[(i,j)] is the free space of segment [Qj,Qj+1] from point Pj

    Parameters
    ----------
    param P : px2 numpy_array, Trajectory P
    param Q : qx2 numpy_array, Trajectory Q
    param p : float, number of points in Trajectory P
    param q : float, number of points in Trajectory Q
    param eps : float, reachability distance

    Returns
    -------
    LF : dict, free spaces of segments of P from points of Q
    BF : dict, free spaces of segments of Q from points of P
    """

    cdef dict LF,BF
    cdef int j,i
    cdef double Q_j0,Q_j1,Q_j10,Q_j11,P_i0,P_i1,P_i10,P_i11

    LF={}
    for j from 0 <= j < q:
        for i from 0 <= i < p-1 :
            Q_j0=Q[j,0]
            Q_j1=Q[j,1]
            P_i0=P[i,0]
            P_i1=P[i,1]
            P_i10=P[i+1,0]
            P_i11=P[i+1,1]

            LF.update({(i,j):_free_line(Q_j0,Q_j1,eps,P_i0,P_i1,P_i10,P_i11)})
    BF={}
    for j from 0 <= j < q-1:
        for i from 0 <= i < p :
            Q_j0=Q[j,0]
            Q_j1=Q[j,1]
            Q_j10=Q[j+1,0]
            Q_j11=Q[j+1,1]
            P_i0=P[i,0]
            P_i1=P[i,1]

            BF.update({(i,j):_free_line(P_i0,P_i1,eps,Q_j0,Q_j1,Q_j10,Q_j11)})
    return LF,BF

def c_compute_LF_BF(np.ndarray[np.float64_t,ndim=2] P, np.ndarray[np.float64_t,ndim=2] Q, int p, int q,
double eps):
    """
    Usage
    -----
    Compute all the free space on the boundary of cells in the diagram for polygonal chains P and Q and the given eps
    LF[(i,j)] is the free space of segments [Pi,Pi+1] from point  Qj
    BF[(i,j)] is the free space of segment [Qj,Qj+1] from point Pj

    Parameters
    ----------
    param P : px2 numpy_array, Trajectory P
    param Q : qx2 numpy_array, Trajectory Q
    param p : float, number of points in Trajectory P
    param q : float, number of points in Trajectory Q
    param eps : float, reachability distance

    Returns
    -------
    LF : dict, free spaces of segments of P from points of Q
    BF : dict, free spaces of segments of Q from points of P
    """

    cdef dict LF,BF
    cdef int j,i
    cdef double Q_j0,Q_j1,Q_j10,Q_j11,P_i0,P_i1,P_i10,P_i11

    LF={}
    for j from 0 <= j < q:
        for i from 0 <= i < p-1 :
            Q_j0=Q[j,0]
            Q_j1=Q[j,1]
            P_i0=P[i,0]
            P_i1=P[i,1]
            P_i10=P[i+1,0]
            P_i11=P[i+1,1]

            LF.update({(i,j):_free_line(Q_j0,Q_j1,eps,P_i0,P_i1,P_i10,P_i11)})
    BF={}
    for j from 0 <= j < q-1:
        for i from 0 <= i < p:
            Q_j0=Q[j,0]
            Q_j1=Q[j,1]
            Q_j10=Q[j+1,0]
            Q_j11=Q[j+1,1]
            P_i0=P[i,0]
            P_i1=P[i,1]

            BF.update({(i,j):_free_line(P_i0,P_i1,eps,Q_j0,Q_j1,Q_j10,Q_j11)})
    return LF,BF

cdef _compute_LR_BR(dict LF, dict BF,int p, int q):
    """
    Usage
    -----
    Compute all the free space,that are reachable from the origin (P[0,0],Q[0,0]) on the boundary of cells
    in the diagram for polygonal chains P and Q and the given free spaces LR and BR

    LR[(i,j)] is the free space, reachable from the origin, of segment [Pi,Pi+1] from point  Qj
    BR[(i,j)] is the free space, reachable from the origin, of segment [Qj,Qj+1] from point Pj

    Parameters
    ----------
    LF : dict, free spaces of segments of P from points of Q
    BF : dict, free spaces of segments of Q from points of P
    param p : float, number of points in Trajectory P
    param q : float, number of points in Trajectory Q

    Returns
    -------
    rep : bool, return true if frechet distance is inf to eps
    LR : dict, is the free space, reachable from the origin, of segments of P from points of Q
    BR : dict, is the free space, reachable from the origin, of segments of Q from points of P
    """

    cdef int i,j
    cdef dict LR,BR
    cdef bool rep

    if not(LF[(0,0)][0] <=0 and BF[(0,0)][0] <=0 and LF[(p-2,q-1)][1]  >=1 and BF[(p-1,q-2)][1]  >=1 )  :
        rep = False
        LR = {}
        BR = {}
    else:
        LR = {(0,0):True}
        BR = {(0,0):True}
        for i from 1 <= i < p-1:

            if (LF[(i,0)][0]!=-1.0 or LF[(i,0)][1]!=-1) and (LF[(i-1,0)][0]==0 and LF[(i-1,0)][1]==1):
                LR[(i,0)]=True
            else:
                LR[(i,0)]=False
        for j from 1 <= j < q-1:
            if (BF[(0,j)][0]!=-1.0 or BF[(0,j)][1]!=-1.0) and (BF[(0,j-1)][0]==0 and BF[(0,j-1)][1]==1):
                BR[(0,j)]=True
            else:
                BR[(0,j)]=False
        for i from 0 <= i < p-1:
            for j from 0 <= j < q-1:
                if LR[(i,j)] or BR[(i,j)]:
                    if LF[(i,j+1)][0]!= -1.0 or LF[(i,j+1)][1]!=-1.0:
                        LR[(i,j+1)]=True
                    else:
                        LR[(i,j+1)]=False
                    if  BF[(i+1,j)][0]!=-1.0 or BF[(i+1,j)][1]!=-1.0:
                        BR[(i+1,j)]=True
                    else:
                        BR[(i+1,j)]=False
                else:
                    LR[(i,j+1)]=False
                    BR[(i+1,j)]=False
        rep = BR[(p-2,q-2)] or LR[(p-2,q-2)]
    return rep,LR,BR

def c_compute_LR_BR(dict LF, dict BF,int p, int q):
    """
    Usage
    -----
    Compute all the free space,that are reachable from the origin (P[0,0],Q[0,0]) on the boundary of cells
    in the diagram for polygonal chains P and Q and the given free spaces LR and BR

    LR[(i,j)] is the free space, reachable from the origin, of segment [Pi,Pi+1] from point  Qj
    BR[(i,j)] is the free space, reachable from the origin, of segment [Qj,Qj+1] from point Pj

    Parameters
    ----------
    LF : dict, free spaces of segments of P from points of Q
    BF : dict, free spaces of segments of Q from points of P
    param p : float, number of points in Trajectory P
    param q : float, number of points in Trajectory Q

    Returns
    -------
    rep : bool, return true if frechet distance is inf to eps
    LR : dict, is the free space, reachable from the origin, of segments of P from points of Q
    BR : dict, is the free space, reachable from the origin, of segments of Q from points of P
    """

    cdef int i,j
    cdef dict LR,BR
    cdef bool rep

    if not(LF[(0,0)][0] <=0 and BF[(0,0)][0] <=0 and LF[(p-2,q-1)][1]  >=1 and BF[(p-1,q-2)][1]  >=1 )  :
        rep = False
        LR={}
        BR={}
    else:
        LR = {(0,0):True}
        BR = {(0,0):True}
        for i from 0 <= i < p-1:

            if (LF[(i,0)][0]!=-1.0 or LF[(i,0)][1]!=-1) and (LF[(i-1,0)][0]==0 and LF[(i-1,0)][1]==1):
                LR[(i,0)]=True
            else:
                LR[(i,0)]=False
        for j from 1 <= j < q-1:
            if (BF[(0,j)][0]!=-1.0 or BF[(0,j)][1]!=-1.0) and (BF[(0,j-1)][0]==0 and BF[(0,j-1)][1]==1):
                BR[(0,j)]=True
            else:
                BR[(0,j)]=False
        for i from 0 <= i < p-1:
            for j from 0 <= j < q-1:
                if LR[(i,j)] or BR[(i,j)]:
                    if LF[(i,j+1)][0]!= -1.0 or LF[(i,j+1)][1]!=-1.0:
                        LR[(i,j+1)]=True
                    else:
                        LR[(i,j+1)]=False
                    if  BF[(i+1,j)][0]!=-1.0 or BF[(i+1,j)][1]!=-1.0:
                        BR[(i+1,j)]=True
                    else:
                        BR[(i+1,j)]=False
                else:
                    LR[(i,j+1)]=False
                    BR[(i+1,j)]=False
        rep = BR[(p-2,q-2)] or LR[(p-2,q-2)]
    return rep,LR,BR


cdef bool _decision_problem(np.ndarray[np.float64_t,ndim=2] P, np.ndarray[np.float64_t,ndim=2] Q,int p,int q, double
eps):
    """
    Usage
    -----
    Test is the frechet distance between trajectories P and Q are inferior to eps

    Parameters
    ----------
    param P : px2 numpy_array, Trajectory P
    param Q : qx2 numpy_array, Trajectory Q
    param p : float, number of points in Trajectory P
    param q : float, number of points in Trajectory Q
    param eps : float, reachability distance

    Returns
    -------
    rep : bool, return true if frechet distance is inf to eps
    """
    cdef dict LF,BF
    cdef bool rep

    LF,BF= _compute_LF_BF(P,Q,p,q,eps)
    rep,_,_ =_compute_LR_BR(LF,BF,p,q)
    return rep

def c_decision_problem(np.ndarray[np.float64_t,ndim=2] P, np.ndarray[np.float64_t,ndim=2] Q,int p,int q, double eps):
    """
    Usage
    -----
    Test is the frechet distance between trajectories P and Q are inferior to eps

    Parameters
    ----------
    param P : px2 numpy_array, Trajectory P
    param Q : qx2 numpy_array, Trajectory Q
    param p : float, number of points in Trajectory P
    param q : float, number of points in Trajectory Q
    param eps : float, reachability distance

    Returns
    -------
    rep : bool, return true if frechet distance is inf to eps
    """
    cdef dict LF,BF
    cdef bool rep

    LF,BF= _compute_LF_BF(P,Q,p,q,eps)
    rep,_,_ =_compute_LR_BR(LF,BF,p,q)
    return rep

cdef list _compute_critical_values(np.ndarray[np.float64_t,ndim=2] P, np.ndarray[np.float64_t,ndim=2] Q,int p,int q):
    """
    Usage
    -----
    Compute all the critical values between trajectories P and Q

    Parameters
    ----------
    param P : px2 numpy_array, Trajectory P
    param Q : qx2 numpy_array, Trajectory Q
    param p : float, number of points in Trajectory P
    param q : float, number of points in Trajectory Q

    Returns
    -------
    cc : list, all critical values between trajectories P and Q
    """
    cdef double origin,end,end_point,Lij,Bij
    cdef set cc
    cdef int i,j

    origin = c_eucl_dist(P[0,0],P[0,1],Q[0,0],Q[0,1])
    end = c_eucl_dist(P[p-1,0],P[p-1,1],Q[q-1,0],Q[q-1,1])
    end_point=fmax(origin,end)
    cc=set([end_point])
    for i from 0 <= i < p-1:
        for j from 0 <= j < q-1:
            Lij=c_point_to_seg(Q[j,0],Q[j,1],P[i,0],P[i,1],P[i+1,0],P[i+1,1])
            if Lij>end_point:
                cc.add(Lij)
            Bij=c_point_to_seg(P[i,0],P[i,1],Q[j,0],Q[j,1],Q[j+1,0],Q[j+1,1])
            if Bij>end_point:
                cc.add(Bij)
    return sorted(list(cc))

def c_compute_critical_values(np.ndarray[np.float64_t,ndim=2] P, np.ndarray[np.float64_t,ndim=2] Q,int p,int q):
    """
    Usage
    -----
    Compute all the critical values between trajectories P and Q

    Parameters
    ----------
    param P : px2 numpy_array, Trajectory P
    param Q : qx2 numpy_array, Trajectory Q
    param p : float, number of points in Trajectory P
    param q : float, number of points in Trajectory Q

    Returns
    -------
    cc : list, all critical values between trajectories P and Q
    """

    cdef double origin,end,end_point,Lij,Bij
    cdef set cc
    cdef int i,j

    origin = c_eucl_dist(P[0,0],P[0,1],Q[0,0],Q[0,1])
    end = c_eucl_dist(P[p-1,0],P[p-1,1],Q[q-1,0],Q[q-1,1])
    end_point=fmax(origin,end)
    cc=set([end_point])
    for i from 0 <= i < p-1:
        for j from 0 <= j < q-1:
            Lij=c_point_to_seg(Q[j,0],Q[j,1],P[i,0],P[i,1],P[i+1,0],P[i+1,1])
            if Lij>end_point:
                cc.add(Lij)
            Bij=c_point_to_seg(P[i,0],P[i,1],Q[j,0],Q[j,1],Q[j+1,0],Q[j+1,1])
            if Bij>end_point:
                cc.add(Bij)
    return sorted(list(cc))

def c_frechet(np.ndarray[np.float64_t,ndim=2] P,np.ndarray[np.float64_t,ndim=2] Q):
    """
    Usage
    -----
    Compute the frechet distance between trajectories P and Q

    Parameters
    ----------
    param P : px2 numpy_array, Trajectory P
    param Q : qx2 numpy_array, Trajectory Q

    Returns
    -------
    frech : float, the frechet distance between trajectories P and Q
    """

    cdef int p,q,m_i
    cdef double eps
    cdef bool rep
    cdef list cc
    p=len(P)
    q=len(Q)

    cc=_compute_critical_values(P,Q,p,q)
    eps=cc[0]
    while(len(cc)!=1):
        m_i=len(cc)/2-1
        eps = cc[m_i]
        rep = _decision_problem(P,Q,p,q,eps)
        if rep:
            cc=cc[:m_i+1]
        else:
            cc=cc[m_i+1:]
    return eps






import numpy as np
from .basic_euclidean import eucl_dist, eucl_dist_traj, point_to_trajectory
from .basic_spherical import point_to_path, great_circle_distance, great_circle_distance_traj


######################
# Euclidean Geometry #
######################

def e_spd(t1, t2, mdist, l_t1, l_t2, t2_dist):
    """
    Usage
    -----
    The spd-distance of trajectory t2 from trajectory t1
    The spd-distance is the sum of the all the point-to-trajectory distance of points of t1 from trajectory t2

    Parameters
    ----------
    param t1 :  l_t1 x 2 numpy_array
    param t2 :  l_t2 x 2 numpy_array
    mdist : len(t1) x len(t2) numpy array, pairwise distance between points of trajectories t1 and t2
    param l_t1: int, length of t1
    param l_t2: int, length of t2
    param t2_dist:  l_t1 x 1 numpy_array,  distances between consecutive points in t2


    Returns
    -------
    spd : float
           spd-distance of trajectory t2 from trajectory t1
    """

    spd = sum([point_to_trajectory(t1[i1], t2, mdist[i1], t2_dist, l_t2) for i1 in range(l_t1)]) / l_t1
    return spd


def e_sspd(t1, t2):
    """
    Usage
    -----
    The sspd-distance between trajectories t1 and t2.
    The sspd-distance isjthe mean of the spd-distance between of t1 from t2 and the spd-distance of t2 from t1.

    Parameters
    ----------
    param t1 :  len(t1)x2 numpy_array
    param t2 :  len(t2)x2 numpy_array

    Returns
    -------
    sspd : float
            sspd-distance of trajectory t2 from trajectory t1
    """
    mdist = eucl_dist_traj(t1, t2)
    l_t1 = len(t1)
    l_t2 = len(t2)
    t1_dist = [eucl_dist(t1[it1], t1[it1 + 1]) for it1 in range(l_t1 - 1)]
    t2_dist = [eucl_dist(t2[it2], t2[it2 + 1]) for it2 in range(l_t2 - 1)]

    sspd = (e_spd(t1, t2, mdist, l_t1, l_t2, t2_dist) + e_spd(t2, t1, mdist.T, l_t2, l_t1, t1_dist)) / 2
    return sspd


######################
# Spherical Geometry #
######################

def s_spd(lons0, lats0, lons1, lats1, n0, n1, mdist, t0_dist):
    """
    Usage
    -----
    The spd-distance of trajectory t1 from trajectory t0
    The spd-distance is the sum of the all the point-to-path distance of points of t0 from trajectory t1

    Parameters
    ----------
    param lons0 :  n0 x 1 numpy_array, longitudes of trajectories t0
    param lats0 :  n0 x 1 numpy_array, lattitudes of trajectories t0
    param lons1 :  n1 x 1 numpy_array, longitudes of trajectories t1
    param lats1 :  n1 x 1 numpy_array, lattitudes of trajectories t1
    param n0: int, length of lons0 and lats0
    param n1: int, length of lons1 and lats1
    mdist : len(t0) x len(t1) numpy array, pairwise distance between points of trajectories t0 and t1
    param t0_dist:  l_t1 x 1 numpy_array,  distances between consecutive points in t0

    Returns
    -------
    spd : float
           spd-distance of trajectory t2 from trajectory t1
    """

    dist = 0
    for j in range(n1):
        dist_j0 = 9e100
        for i in range(n0 - 1):
            dist_j0 = np.min((dist_j0, point_to_path(lons0[i], lats0[i], lons0[i + 1], lats0[i + 1], lons1[j],
                                                     lats1[j], mdist[i, j], mdist[i + 1, j], t0_dist[i])))
        dist = dist + dist_j0
    dist = float(dist) / n1
    return dist


def s_sspd(t0, t1):
    """
    Usage
    -----
    The sspd-distance between trajectories t1 and t2.
    The sspd-distance is the mean of the spd-distance between of t1 from t2 and the spd-distance of t2 from t1.

    Parameters
    ----------
    param t0 :  len(t0)x2 numpy_array
    param t1 :  len(t1)x2 numpy_array

    Returns
    -------
    sspd : float
            sspd-distance of trajectory t2 from trajectory t1
    """
    n0 = len(t0)
    n1 = len(t1)
    lats0 = t0[:, 1]
    lons0 = t0[:, 0]
    lats1 = t1[:, 1]
    lons1 = t1[:, 0]

    mdist = great_circle_distance_traj(lons0, lats0, lons1, lats1, n0, n1)

    t0_dist = [great_circle_distance(lons0[it0], lats0[it0], lons0[it0 + 1], lats0[it0 + 1]) for it0 in range(n0 - 1)]
    t1_dist = [great_circle_distance(lons1[it1], lats1[it1], lons1[it1 + 1], lats1[it1 + 1]) for it1 in range(n1 - 1)]

    dist = s_spd(lons0, lats0, lons1, lats1, n0, n1, mdist, t0_dist) + s_spd(lons1, lats1, lons0, lats0, n1, n0,
                                                                             mdist.T, t1_dist)
    return dist



import pandas as pd
import pickle


data = pd.read_pickle("/Users/bguillouet/These/trajectory_review/data/extracted/starting_from/Caltrain_city_center_v3.pkl")

traj_list = [group[["lons","lats"]].values for _,group in data.groupby("id_traj")]

pickle.dump(traj_list, open("/Users/bguillouet/These/trajectory_distance/data/" + "benchmark_trajectories.pkl","wb"))


import math
import numpy as np

rad = math.pi / 180.0
R = 6378137.0


def spherical2Cart(lon, lat):
    clat=(90-lat)*rad
    lon=lon*rad
    x=math.cos(lon)*math.sin(clat)
    y=math.sin(lon)*math.sin(clat)
    z=math.cos(clat)

    return [x,y,z]

def cart2Spherical(x,y,z):
    r=math.sqrt(x**2+y**2+z**2)
    clat=math.acos(z/r)/math.pi*180
    lat=90.-clat
    lon=math.atan2(y,x)/math.pi*180
    lon=(lon+360)%360

    return [lon, lat]

def great_circle_distance(lon1, lat1, lon2, lat2):
    """
    Usage
    -----
    Compute the great circle distance, in meter, between (lon1,lat1) and (lon2,lat2)

    Parameters
    ----------
    param lat1: float, latitude of the first point
    param lon1: float, longitude of the first point
    param lat2: float, latitude of the se*cond point
    param lon2: float, longitude of the second point

    Returns
    -------x
    d: float
       Great circle distance between (lon1,lat1) and (lon2,lat2)
    """

    dlat = rad * (lat2 - lat1)
    dlon = rad * (lon2 - lon1)
    a = (math.sin(dlat / 2.0) * math.sin(dlat / 2.0) +
         math.cos(rad * lat1) * math.cos(rad * lat2) *
         math.sin(dlon / 2.0) * math.sin(dlon / 2.0))
    c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
    d = R * c
    return d


def great_circle_distance_traj(lons1, lats1, lons2, lats2, l1, l2):
    """
    Usage
    -----
    Compute pairwise great circle distance, in meter, between longitude/latitudes coordinates.

    Parameters
    ----------
    param lats1: float, latitudes of the firs trajectories
    param lons1: float, longitude of the trajectories
    param lats2: float, latitudes of the se*cond trajectories
    param lons2: float, longitudess of the second trajectories
    param l1 : int, length of the first trajectories
    param l2 : int, length of the second trajectories

    Returns
    -------x
    d: float
       Great circle distance between (lon1,lat1) and (lon2,lat2)
    """

    mdist = np.empty((l1, l2), dtype=float)
    for i in range(l1):
        for j in range(l2):
            mdist[i, j] = great_circle_distance(lons1[i], lats1[i], lons2[j], lats2[j])
    return mdist


def initial_bearing(lon1, lat1, lon2, lat2):
    """
    Usage
    -----
    Bearing between (lon1,lat1) and (lon2,lat2), in degree.

    Parameters
    ----------
    param lat1: float, latitude of the first point
    param lon1: float, longitude of the first point
    param lat2: float, latitude of the second point
    param lon2: float, longitude of the second point

    Returns
    -------
    brng: float
           Bearing between (lon1,lat1) and (lon2,lat2), in degree.

    """
    dlon = rad * (lon2 - lon1)
    y = math.sin(dlon) * math.cos(rad * (lat2))
    x = math.cos(rad * (lat1)) * math.sin(rad * (lat2)) - math.sin(rad * (lat1)) * math.cos(rad * (lat2)) * math.cos(
        dlon)
    ibrng = math.atan2(y, x)

    return ibrng

def cross_track_point(lon1, lat1, lon2, lat2, lon3, lat3):
    '''Get the closest point on great circle path to the 3rd point

    <lat1>, <lon1>: scalar float or nd-array, latitudes and longitudes in
                    degree, start point of the great circle.
    <lat2>, <lon2>: scalar float or nd-array, latitudes and longitudes in
                    degree, end point of the great circle.
    <lat3>, <lon3>: scalar float or nd-array, latitudes and longitudes in
                    degree, a point away from the great circle.

    Return <latp>, <lonp>: latitude and longitude of point P on the great
                           circle that connects P1, P2, and is closest
                           to point P3.
    '''

    x1,y1,z1=spherical2Cart(lon1,lat1)
    x2,y2,z2=spherical2Cart(lon2,lat2)
    x3,y3,z3=spherical2Cart(lon3,lat3)

    D,E,F=np.cross([x1,y1,z1],[x2,y2,z2])

    a=E*z3-F*y3
    b=F*x3-D*z3
    c=D*y3-E*x3

    f=c*E-b*F
    g=a*F-c*D
    h=b*D-a*E

    tt=math.sqrt(f**2+g**2+h**2)
    xp=f/tt
    yp=g/tt
    zp=h/tt

    lon1, lat1 =cart2Spherical(xp,yp,zp)
    lon2, lat2 =cart2Spherical(-xp,-yp,-zp)
    #TODO MIGHT REQUIRE EARTH RADIUS  https://gis.stackexchange.com/questions/209540/projecting-cross-track-distance-on-great-circle
    d1=great_circle_distance(lon1, lat1, lon3, lat3)
    d2=great_circle_distance(lon2, lat2, lon3, lat3)

    if d1>d2:
        return lon2, lat2
    else:
        return lon1, lat1

def cross_track_distance(lon1, lat1, lon2, lat2, lon3, lat3, d13):
    """
    Usage
    -----
    Angular cross-track-distance of a point (lon3, lat3) from a great-circle path between (lon1, lat1) and (lon2, lat2)
    The sign of this distance tells which side of the path the third point is on.

    Parameters :
    ----------
    param lat1: float, latitude of the first point
    param lon1: float, longitude of the first point
    param lat2: float, latitude of the second point
    param lon2: float, longitude of the second point
     param lat3: float, latitude of the third point
    param lon3: float, longitude of the third point

    Usage
    -----
    crt: float
         the (angular) cross_track_distance

    """

    theta13 = initial_bearing(lon1, lat1, lon3, lat3)  # bearing from start point to third point
    theta12 = initial_bearing(lon1, lat1, lon2, lat2)  # bearing from start point to end point

    crt = math.asin(math.sin(d13 / R) * math.sin(theta13 - theta12)) * R

    return crt


def along_track_distance(crt, d13):
    """
    Usage
    -----
    The along-track distance from the start point (lon1, lat1) to the closest point on the the path
    to the third point (lon3, lat3).

    Parameters
    ----------
    param crt : float, cross_track_distance
    param d13 : float, along_track_distance

    Returns
    -------
    alt: float
         The along-track distance
    """

    alt = math.acos(math.cos(d13 / R) / math.cos(crt / R)) * R
    return alt


def point_to_path(lon1, lat1, lon2, lat2, lon3, lat3, d13, d23, d12):
    """
    Usage
    -----
    The point-to-path distance between point (lon3, lat3) and path delimited by (lon1, lat1) and (lon2, lat2).
    The point-to-path distance is the cross_track distance between the great circle path if the projection of
    the third point lies on the path. If it is not on the path, return the minimum of the
    great_circle_distance between the first and the third or the second and the third point.

    Parameters
    ----------
    param lat1: float, latitude of the first point
    param lon1: float, longitude of the first point
    param lat2: float, latitude of the second point
    param lon2: float, longitude of the second point
    param lat3: float, latitude of the third point
    param lon3: float, longitude of the third point
    param d13 : float, great circle distance between (lon1, lat1) and (lon3, lat3)
    param d23 : float, great circle distance between (lon2, lat2) and (lon3, lat3)
    param d12 : float, great circle distance between (lon1, lat1) and (lon2, lat2)


    Returns
    -------

    ptp : float
          The point-to-path distance between point (lon3, lat3) and path delimited by (lon1, lat1) and (lon2, lat2)

    """
    crt = cross_track_distance(lon1, lat1, lon2, lat2, lon3, lat3, d13)
    d1p = along_track_distance(crt, d13)
    d2p = along_track_distance(crt, d23)
    if (d1p > d12) or (d2p > d12):
        ptp = np.min((d13, d23))
    else:
        ptp = np.abs(crt)
    return ptp



from .basic_euclidean import point_to_trajectory, eucl_dist_traj, eucl_dist
from .basic_spherical import point_to_path, great_circle_distance, great_circle_distance_traj


######################
# Euclidean Geometry #
######################

def e_directed_hausdorff(t1, t2, mdist, l_t1, l_t2, t2_dist):
    """
    Usage
    -----
    directed hausdorff distance from trajectory t1 to trajectory t2.

    Parameters
    ----------
    Parameters
    ----------
    param t1 :  l_t1 x 2 numpy_array
    param t2 :  l_t2 x 2 numpy_array
    mdist : len(t1) x len(t2) numpy array, pairwise distance between points of trajectories t1 and t2
    param l_t1: int, length of t1
    param l_t2: int, length of t2
    param t2_dist:  l_t1 x 1 numpy_array,  distances between consecutive points in t2



    Returns
    -------
    dh : float, directed hausdorff from trajectory t1 to trajectory t2
    """
    dh = max([point_to_trajectory(t1[i1], t2, mdist[i1], t2_dist, l_t2) for i1 in range(l_t1)])
    return dh


def e_hausdorff(t1, t2):
    """
    Usage
    -----
    hausdorff distance between trajectories t1 and t2.

    Parameters
    ----------
    param t1 :  len(t1)x2 numpy_array
    param t2 :  len(t2)x2 numpy_array

    Returns
    -------
    h : float, hausdorff from trajectories t1 and t2
    """
    mdist = eucl_dist_traj(t1, t2)
    l_t1 = len(t1)
    l_t2 = len(t2)
    t1_dist = [eucl_dist(t1[it1], t1[it1 + 1]) for it1 in range(l_t1 - 1)]
    t2_dist = [eucl_dist(t2[it2], t2[it2 + 1]) for it2 in range(l_t2 - 1)]

    h = max(e_directed_hausdorff(t1, t2, mdist, l_t1, l_t2, t2_dist),
            e_directed_hausdorff(t2, t1, mdist.T, l_t2, l_t1, t1_dist))
    return h


######################
# Spherical Geometry #
######################

def s_directed_hausdorff(lons0, lats0, lons1, lats1, n0, n1, mdist, t0_dist):
    """
    Usage
    -----
    directed hausdorff distance from trajectory t1 to trajectory t2.

    Parameters
    ----------
    param t1 :  len(t1)x2 numpy_array
    param t2 :  len(t2)x2 numpy_array

    Returns
    -------
    dh : float, directed hausdorff from trajectory t1 to trajectory t2
    """

    dh = 0
    for j in range(n1):
        dist_j0 = 9e100
        for i in range(n0 - 1):
            dist_j0 = min(dist_j0, point_to_path(lons0[i], lats0[i], lons0[i + 1], lats0[i + 1], lons1[j],
                                                 lats1[j], mdist[i, j], mdist[i + 1, j], t0_dist[i]))
        dh = max(dh, dist_j0)
    return dh


def s_hausdorff(t0, t1):
    """
    Usage
    -----
    hausdorff distance between trajectories t1 and t2.

    Parameters
    ----------
    param lons0 :  n0 x 1 numpy_array, longitudes of trajectories t0
    param lats0 :  n0 x 1 numpy_array, lattitudes of trajectories t0
    param lons1 :  n1 x 1 numpy_array, longitudes of trajectories t1
    param lats1 :  n1 x 1 numpy_array, lattitudes of trajectories t1
    param n0: int, length of lons0 and lats0
    param n1: int, length of lons1 and lats1
    mdist : len(t0) x len(t1) numpy array, pairwise distance between points of trajectories t0 and t1
    param t0_dist:  l_t1 x 1 numpy_array,  distances between consecutive points in t0


    Returns
    -------
    h : float, hausdorff from trajectories t1 and t2
    """
    n0 = len(t0)
    n1 = len(t1)
    lats0 = t0[:, 1]
    lons0 = t0[:, 0]
    lats1 = t1[:, 1]
    lons1 = t1[:, 0]

    mdist = great_circle_distance_traj(lons0, lats0, lons1, lats1, n0, n1)

    t0_dist = [great_circle_distance(lons0[it0], lats0[it0], lons0[it0 + 1], lats0[it0 + 1]) for it0 in range(n0 - 1)]
    t1_dist = [great_circle_distance(lons1[it1], lats1[it1], lons1[it1 + 1], lats1[it1 + 1]) for it1 in range(n1 - 1)]

    h = max(s_directed_hausdorff(lons0, lats0, lons1, lats1, n0, n1, mdist, t0_dist),
            s_directed_hausdorff(lons1, lats1, lons0, lats0, n1, n0, mdist.T, t1_dist))
    return h
